import historydag.compact_genome as compact_genome
from historydag.compact_genome import CompactGenome
import historydag.parsimony_utils
from historydag.parsimony_utils import default_nt_transitions


def test_sequence_cg_convert():
    seqs = [
        "AAAA",
        "TAAT",
        "CTGA",
        "TGCA",
    ]
    for refseq in seqs:
        for seq in seqs:
            cg = compact_genome.compact_genome_from_sequence(seq, refseq)
            reseq = cg.to_sequence()
            if reseq != seq:
                print("\nUnmatched reconstructed sequence:")
                print("ref sequence:", refseq)
                print("sequence:", seq)
                print("cg:", cg)
                print("reconstructed sequence:", reseq)
                assert False


def test_cg_diff():
    refseq = "C" * 1000
    cgs = [
        CompactGenome({287: ("C", "G")}, refseq),
        CompactGenome({287: ("C", "G"), 318: ("C", "A"), 495: ("C", "T")}, refseq),
        CompactGenome(
            {287: ("C", "G"), 80: ("C", "T"), 257: ("C", "G"), 591: ("C", "A")}, refseq
        ),
        CompactGenome(
            {
                287: ("C", "G"),
                191: ("C", "G"),
                492: ("C", "G"),
                612: ("C", "G"),
                654: ("C", "G"),
            },
            refseq,
        ),
        CompactGenome({287: ("C", "G"), 318: ("C", "A"), 495: ("C", "T")}, refseq),
    ]
    # First try with apply_muts
    for parent_cg in cgs:
        for child_cg in cgs:
            assert (
                parent_cg.apply_muts(
                    str(p) + str(key) + str(c)
                    for p, c, key in compact_genome.cg_diff(parent_cg, child_cg)
                )
                == child_cg
            )
            assert all(
                par_nuc != child_nuc
                for par_nuc, child_nuc, idx in compact_genome.cg_diff(
                    parent_cg, child_cg
                )
            )
    # Now with apply_muts_raw
    for parent_cg in cgs:
        for child_cg in cgs:
            assert (
                parent_cg.apply_muts_raw(
                    (key, p, c)
                    for p, c, key in compact_genome.cg_diff(parent_cg, child_cg)
                )
                == child_cg
            )
            assert all(
                par_nuc != child_nuc
                for par_nuc, child_nuc, idx in compact_genome.cg_diff(
                    parent_cg, child_cg
                )
            )


def test_ambiguous_cg_distance():
    reference_seq = "AAAAAAAN"

    s1 = compact_genome.CompactGenome({1: ("A", "T"), 2: ("A", "G")}, reference_seq)
    s2 = compact_genome.CompactGenome({1: ("A", "T"), 2: ("A", "N")}, reference_seq)
    s3 = compact_genome.CompactGenome({1: ("A", "T")}, reference_seq)
    s4 = compact_genome.CompactGenome({8: ("N", "C")}, reference_seq)

    assert default_nt_transitions.min_weighted_cg_hamming_distance(s1, s2) == 0
    assert default_nt_transitions.min_weighted_cg_hamming_distance(s1, s3) == 1
    assert default_nt_transitions.min_weighted_cg_hamming_distance(s1, s4) == 3
    assert default_nt_transitions.min_weighted_cg_hamming_distance(s4, s1) == 2
    assert default_nt_transitions.min_weighted_cg_hamming_distance(s3, s2) == 0
    assert default_nt_transitions.min_weighted_cg_hamming_distance(s4, s2) == 1
    assert default_nt_transitions.min_weighted_cg_hamming_distance(s4, s3) == 1


def test_reconcile_cgs():
    # ambiguitymap = historydag.parsimony_utils.standard_nt_ambiguity_map_gap_as_char
    ambiguitymap = historydag.parsimony_utils.standard_nt_ambiguity_map
    reference_seq = "AAAAAA-N"

    s1 = compact_genome.CompactGenome({1: ("A", "T"), 2: ("A", "G")}, reference_seq)
    s2 = compact_genome.CompactGenome({1: ("A", "C"), 2: ("A", "G")}, reference_seq)
    s3 = compact_genome.CompactGenome({1: ("A", "T")}, reference_seq)
    s4 = compact_genome.CompactGenome({8: ("N", "C")}, reference_seq)
    s5 = compact_genome.CompactGenome({1: ("A", "G"), 2: ("A", "T")}, reference_seq)

    a1 = compact_genome.CompactGenome({1: ("A", "Y"), 2: ("A", "G")}, reference_seq)
    a2 = compact_genome.CompactGenome({1: ("A", "Y"), 2: ("A", "R")}, reference_seq)
    a3 = compact_genome.CompactGenome({1: ("A", "B"), 2: ("A", "K")}, reference_seq)
    a4 = compact_genome.CompactGenome({1: ("A", "R"), 2: ("A", "W")}, reference_seq)
    a5 = compact_genome.CompactGenome({1: ("A", "D"), 2: ("A", "W")}, reference_seq)

    seqs = [s1, s2, s3, s4, s5]

    for s in seqs:
        assert s == compact_genome.reconcile_cgs((s,), ambiguitymap=ambiguitymap)[0]

    testcases = [
        ((s1, s2), a1),
        ((s1, s2, s3), a2),
        ((s1, s2, s5), a3),
        ((s4, s5), a4),
        ((s3, s4, s5), a5),
        ((s1, s1), s1),
    ]

    for cgs, answer in testcases:
        assert compact_genome.reconcile_cgs(cgs, ambiguitymap=ambiguitymap)[0] == answer


def test_ambiguous_cg_diff():
    ref = "AAAAAA"
    c1 = compact_genome.CompactGenome({}, ref)
    c2 = compact_genome.CompactGenome(
        {1: ("A", "C"), 2: ("A", "N"), 4: ("A", "K")}, ref
    )
    muts = set(compact_genome.ambiguous_cg_diff(c1, c2))
    assert muts == {("A", "C", 1), ("A", "G", 4)} or muts == {
        ("A", "C", 1),
        ("A", "T", 4),
    }



from historydag.dag import (
    HistoryDagNode,
    EdgeSet,
    from_tree,
    history_dag_from_newicks,
    from_newick,
)
from historydag import utils
import ete3
from collections import Counter
import random

""" HistoryDag tests:"""

newickstring1 = (
    "((4[&&NHX:name=4:sequence=K],(6[&&NHX:name=6:sequence=S],"
    "7[&&NHX:name=7:sequence=T])5[&&NHX:name=5:sequence=M])3[&&NHX:name=3:sequence=M],"
    "8[&&NHX:name=8:sequence=W],(11[&&NHX:name=11:sequence=V],"
    "10[&&NHX:name=10:sequence=D])9[&&NHX:name=9:sequence=R])1[&&NHX:name=1:sequence=G];"
)

newickstring2 = (
    "((4[&&NHX:name=4:sequence=K],(6[&&NHX:name=6:sequence=S],"
    "7[&&NHX:name=7:sequence=T])5[&&NHX:name=5:sequence=H])3[&&NHX:name=3:sequence=G],"
    "8[&&NHX:name=8:sequence=W],(11[&&NHX:name=11:sequence=V],"
    "10[&&NHX:name=10:sequence=D])9[&&NHX:name=9:sequence=C])1[&&NHX:name=1:sequence=A];"
)

newickstring3 = (
    "((4[&&NHX:name=4:sequence=K],(6[&&NHX:name=6:sequence=S],"
    "7[&&NHX:name=7:sequence=T])5[&&NHX:name=5:sequence=H])2[&&NHX:name=2:sequence=B],"
    "8[&&NHX:name=8:sequence=W],(11[&&NHX:name=11:sequence=V],"
    "10[&&NHX:name=10:sequence=D])9[&&NHX:name=9:sequence=C])1[&&NHX:name=1:sequence=A];"
)

namedict = {
    "A": 1,
    "B": 2,
    "C": 9,
    "D": 10,
    "V": 11,
    "W": 8,
    "G": 3,
    "H": 5,
    "T": 7,
    "S": 6,
    "K": 4,
    None: "DAG_root",
}


def test_init():
    r = HistoryDagNode(("a",), {}, None)
    assert r.is_leaf()
    r = HistoryDagNode(
        ("a",),
        {frozenset(["a", "b"]): EdgeSet(), frozenset(["c", "d"]): EdgeSet()},
        None,
    )
    assert not r.is_leaf()
    s = HistoryDagNode(
        ("b",),
        {frozenset(["a", "b"]): EdgeSet(), frozenset(["c", "d"]): EdgeSet([r])},
        None,
    )
    assert not s.is_leaf()


def test_edge():
    r = HistoryDagNode(("a",), {}, None)
    r2 = HistoryDagNode(
        ("b",),
        {frozenset({("z",), ("y",)}): EdgeSet(), frozenset({("a",)}): EdgeSet()},
        None,
    )
    s = HistoryDagNode(
        ("b",),
        {frozenset([("a",)]): EdgeSet(), frozenset([("c",), ("d",)]): EdgeSet()},
        None,
    )
    s.add_edge(r)
    try:
        s.add_edge(r2)
        assert False
    except KeyError:
        pass


def test_from_tree():
    tree = ete3.Tree(newickstring2, format=1)
    print(tree.sequence)
    dag = from_tree(tree, ["sequence"])
    dag._check_valid()
    dag.to_graphviz(namedict=namedict)


def test_is_history():
    tree = ete3.Tree(newickstring2, format=1)
    print(tree.sequence)
    dag = from_tree(tree, ["sequence"])
    dag._check_valid()
    assert dag.is_history()


def test_from_tree_label():
    tree = ete3.Tree(newickstring2, format=1)
    for node in tree.traverse():
        node.add_feature("abundance", 1)
    dag = from_tree(  # noqa
        tree,
        ["sequence", "abundance"],
        label_functions={"abundance2x": lambda n: 2 * n.abundance},
    )
    dag._check_valid()


def test_preserve_attr():
    # attr is populated, and preserved by dag operations which add nodes.
    dag = history_dag_from_newicks(
        [newickstring1, newickstring2, newickstring3],
        ["sequence"],
        attr_func=lambda n: n.name,
    )
    dag._check_valid()
    assert all(n.attr for n in dag.preorder(skip_ua_node=True))

    @utils.explode_label("sequence")
    def expand_func(seq):
        if seq == "C":
            yield from ["D", "A"]
        elif seq == "H":
            yield from ["T", "W"]
        else:
            yield seq

    dag.explode_nodes(expand_func=expand_func)
    dag._check_valid()
    assert all(n.attr for n in dag.preorder(skip_ua_node=True))
    dag.convert_to_collapsed()
    dag._check_valid()
    assert all(n.attr for n in dag.preorder(skip_ua_node=True))


def test_ete_newick_agree():
    dag = history_dag_from_newicks(
        [newickstring1, newickstring2, newickstring3],
        ["sequence"],
        attr_func=lambda n: n.name,
    )
    outkwargs = {
        "name_func": lambda n: n.attr,
        "features": ["sequence"],
        "feature_funcs": {},
    }

    inkwargs = {
        "label_features": ["sequence"],
        "label_functions": {},
        "attr_func": lambda n: n.name,
    }
    viaetes = {
        from_tree(tree.to_ete(**outkwargs), **inkwargs).to_newick(**outkwargs)
        for tree in dag.get_histories()
    }
    vianewicks = {tree.to_newick(**outkwargs) for tree in dag.get_histories()}
    assert viaetes == vianewicks


def test_postorder():
    tree = ete3.Tree(newickstring2, format=1)
    dag = from_tree(tree, ["sequence"])
    assert [
        namedict[node.label.sequence]
        if isinstance(node.label, tuple)
        else str(node.label)
        for node in dag.postorder()
    ] == [
        4,
        6,
        7,
        5,
        3,
        8,
        11,
        10,
        9,
        1,
        "UA_Node",
    ]
    # print([namedict[node.label] for node in postorder(dag)])


def test_children():
    tree = ete3.Tree(newickstring2, format=1)
    dag = from_tree(tree, ["sequence"])
    print([child.label for child in dag.dagroot.children()])
    for child in dag.dagroot.children():
        print([cc.label for cc in child.children()])
        for ccc in child.children():
            print([cccc.label for cccc in ccc.children()])


def test_merge():
    tree1 = ete3.Tree(newickstring2, format=1)
    dag1 = from_tree(tree1, ["sequence"])
    tree2 = ete3.Tree(newickstring3, format=1)
    dag2 = from_tree(tree2, ["sequence"])
    dag1.merge(dag2)
    dag1.to_graphviz(namedict=namedict)


def test_weight():
    tree1 = ete3.Tree(newickstring2, format=1)
    dag1 = from_tree(tree1, ["sequence"])
    tree2 = ete3.Tree(newickstring3, format=1)
    dag2 = from_tree(tree2, ["sequence"])
    dag1.merge(dag2)
    dag1.to_graphviz(namedict=namedict)
    assert dag1.optimal_weight_annotate() == 9


def test_internal_avg_parents():
    tree1 = ete3.Tree(newickstring2, format=1)
    dag1 = from_tree(tree1, ["sequence"])
    tree2 = ete3.Tree(newickstring3, format=1)
    dag2 = from_tree(tree2, ["sequence"])
    dag1.merge(dag2)
    dag1.to_graphviz(namedict=namedict)
    assert dag1.internal_avg_parents() == 1.2


def test_sample():
    newicks = ["((a, b)b, c)c;", "((a, b)c, c)c;", "((a, b)a, c)c;", "((a, b)r, c)r;"]
    newicks = ["((1, 2)2, 3)3;", "((1, 2)3, 3)3;", "((1, 2)1, 3)3;", "((1, 2)4, 3)4;"]
    namedict = {(str(x),): x for x in range(5)}
    dag = history_dag_from_newicks(newicks, ["name"])
    for i in range(10):
        assert dag.sample().is_history()
    sample = dag.sample()
    sample._check_valid()
    sample.to_graphviz(namedict=namedict)


def test_unifurcation():
    # Make sure that unifurcations are handled correctly
    # First make sure the call works when the problem is fixed:
    from_newick("((a, b)b, c)c;", ["name"])
    try:
        from_newick("(((a, b)b, c)d)c;", ["name"])
        from_newick("(((a, b)d)b, c)c;", ["name"])
        raise RuntimeError(
            "history DAG was allowed to be constructed from a tree with a unifurcation."
        )
    except ValueError:
        return


def test_unique_leaves():
    # Make sure non-unique leaf labels won't be allowed
    from_newick("((a, b)b, c)c;", ["name"])
    try:
        from_newick("(((a, b)b, a)d)c;", ["name"])
        raise RuntimeError(
            "history DAG creation was allowed with non-unique leaf labels"
        )
    except ValueError:
        return


def test_explode_rejects_leaf_ambiguities():
    # Make sure explode won't expand a leaf
    # First make sure it works if we fix the problem:
    dag = from_newick(
        "((A, C)W, T)C;", [], label_functions={"sequence": lambda n: n.name}
    )
    dag.explode_nodes(expandable_func=None)

    dag = from_newick(
        "((A, N)W, T)C;", [], label_functions={"sequence": lambda n: n.name}
    )
    N = dag.num_leaves()
    dag.explode_nodes(expandable_func=None)
    assert dag._check_valid()
    assert dag.num_leaves() == N


def test_print():
    tree1 = ete3.Tree(newickstring2, format=1)
    dag1 = from_tree(tree1, ["sequence"])
    dag1.__repr__()


def test_eq():
    tree1 = ete3.Tree(newickstring2, format=1)
    dag1 = from_tree(tree1, ["sequence"])
    tree2 = ete3.Tree("((z, b)b, c)c;", format=1)
    dag2 = from_tree(tree2, ["name"])
    assert dag1.dagroot == dag1.copy().dagroot
    assert list(dag1.dagroot.children())[0] != list(dag2.dagroot.children())[0]


def test_to_graphviz():
    dag = from_newick(
        "((aaaaaaaaa, bbbbbbbbb)bbbbbbbbb, ccccccccc)ccccccccc;", ["name"]
    )
    dag.to_graphviz()


def test_reproducible():
    dag = history_dag_from_newicks(
        [newickstring1, newickstring2, newickstring3], ["sequence"]
    )
    dag.make_uniform()
    random.seed(1)
    take1 = [dag.sample().to_newick() for _ in range(1000)]
    random.seed(1)
    take2 = [dag.sample().to_newick() for _ in range(1000)]
    assert len(set(take2)) == 3
    assert take1 == take2


def test_make_uniform():
    random.seed(1)

    def normalize_counts(counter):
        n = len(list(counter.elements()))
        return ([num / n for _, num in counter.items()], (n / len(counter)) / n)

    def is_close(f1, f2):
        return abs(f1 - f2) < 0.03

    dag = history_dag_from_newicks(
        [newickstring1, newickstring2, newickstring3], ["sequence"]
    )
    take1 = Counter([dag.sample().to_newick() for _ in range(1000)])
    dag.make_uniform()
    dag._check_valid()
    take2 = Counter([dag.sample().to_newick() for _ in range(1000)])

    take1norms, avg1 = normalize_counts(take1)
    assert any(not is_close(norm, avg1) for norm in take1norms)

    take2norms, avg2 = normalize_counts(take2)
    assert all(is_close(norm, avg2) for norm in take2norms)

    # now check that sampling with log probabilities works:
    dag.probability_annotate(lambda n1, n2: 0, log_probabilities=True)
    dag._check_valid()
    take2 = Counter(
        [dag.sample(log_probabilities=True).to_newick() for _ in range(1000)]
    )
    take2norms, avg2 = normalize_counts(take2)
    assert all(is_close(norm, avg2) for norm in take2norms)


def test_summary():
    dag = history_dag_from_newicks(
        [newickstring1, newickstring2, newickstring3], ["sequence"]
    )
    dag.summary()


def test_to_newick():
    dag = history_dag_from_newicks(
        [newickstring1, newickstring2, newickstring3], ["sequence"]
    )
    try:
        dag.to_newick()
        raise RuntimeError("to_newick shouldn't accept a DAG that's not a tree")
    except ValueError:
        pass



import ete3
import pickle
import historydag.dag as hdag


def deterministic_newick(tree: ete3.TreeNode) -> str:
    """For use in comparing ete3 TreeNodes with newick strings"""
    newtree = tree.copy()
    for node in newtree.traverse():
        node.name = 1
        node.children.sort(key=lambda node: node.sequence)
        node.dist = 1
    return newtree.write(format=1, features=["sequence"], format_root_node=True)


def deterministic_newick_topology(tree: ete3.TreeNode) -> str:
    """For use in comparing ete3 TreeNodes with newick strings, distinguishing only
    by topologies above leaves."""
    newtree = tree.copy()
    for node in newtree.traverse():
        node.name = node.sequence
        node.children.sort(
            key=lambda node: str(sorted(lf.name for lf in node.get_leaves()))
        )
        node.dist = 1
    return newtree.write(format=9)


newicklistlist = [
    ["((AA, CT)CG, (TA, CC)CG)CC;"],
    ["((AA, CT)CG, (TA, CC)CG)CC;", "((AA, CT)CA, (TA, CC)CC)CC;"],
    [
        "((CA, GG)CA, AA, (TT, (CC, GA)CC)CC)AA;",
        "((CA, GG)CA, AA, (TT, (CC, GA)CA)CA)AA;",
        "((CA, GG)CG, AA, (TT, (CC, GA)GC)GC)AG;",
    ],
    ["((AA, CT)CG, (TA, CC)CG)CC;", "((AA, CT)CA, (TA, CC)CC)CC;"],
    [
        "((CA, GG)CA, AT, (TT, (CC, GA)CC)CC)AA;",
        "((CA, GG)CA, AA, (TT, (CC, GA)CA)CA)AA;",
        "((CA, GG)CG, AA, (TT, (CC, GA)GC)GC)AG;",
    ],
]

dags = [
    hdag.history_dag_from_newicks(
        newicklist, [], label_functions={"sequence": lambda n: n.name}
    )
    for newicklist in newicklistlist
]

with open("sample_data/toy_trees_100_uncollapsed.p", "rb") as fh:
    uncollapsed = pickle.load(fh)
for tree in uncollapsed:
    if len(tree.children) == 1:
        newchild = tree.copy()
        for child in newchild.get_children():
            newchild.remove_child(child)
        tree.add_child(newchild)
        assert newchild.is_leaf()

dags.append(
    hdag.history_dag_from_etes(
        uncollapsed[0:5], [], label_functions={"sequence": lambda n: n.sequence}
    )
)


def test_node_counts():
    print(f"Testing with {len(dags)} dags")
    for dag in dags:
        node2count = dag.count_nodes()
        for node in node2count.keys():
            ground_truth = sum(
                [node in set(tree.postorder()) for tree in dag.get_histories()]
            )
            # print(
            #     f"\t node2count[node] = {node2count[node]} \t ground_truth = {ground_truth}"
            # )
            assert node2count[node] == ground_truth


def test_collapsed_node_counts():
    print(f"Testing with {len(dags)} dags")
    for dag in dags:
        node2count = dag.count_nodes(collapse=True)
        print("new dag")
        # print(dag.to_graphviz())

        for node in node2count.keys():
            ground_truth = sum(
                [
                    node in set([n.clade_union() for n in tree.postorder()])
                    for tree in dag.get_histories()
                ]
            )
            print(
                f"node2count[node] = {node2count[node]} \t ground_truth = {ground_truth}\t {node}"
            )
            assert node2count[node] == ground_truth


# NOTE: Tests that edge2count contains edges -> count that are correct, but not that all
# edges are contained...
def test_edge_counts():
    print(f"Testing with {len(dags)} dags")
    for dag in dags:
        edge2count = dag.count_edges()

        for parent, node in edge2count.keys():
            ground_truth = 0
            for tree in dag.get_histories():
                for curr_parent in tree.postorder():
                    if curr_parent != parent:
                        continue
                    if node in curr_parent.children():
                        ground_truth += 1
                        break

            print(
                f"count={edge2count[(parent, node)]}\tground_truth={ground_truth}\t{parent.label}\t{node.label}"
            )
            print()
            assert edge2count[(parent, node)] == ground_truth



from frozendict import frozendict
from typing import Dict, Sequence
from warnings import warn
from historydag.parsimony_utils import (
    standard_nt_ambiguity_map,
    default_nt_transitions,
    CharacterSequence,
)
from historydag.utils import read_fasta


class CompactGenome:
    """A collection of mutations relative to a reference sequence.

    Args:
        mutations: The difference between the reference and this sequence, expressed
            in a dictionary, in which keys are one-based sequence indices, and values
            are (reference base, new base) pairs.
        reference: The reference sequence
    """

    def __init__(self, mutations: Dict, reference: str):
        self.reference = reference
        self.mutations = frozendict(mutations)

    def __hash__(self):
        return hash(self.mutations)

    def __eq__(self, other):
        if isinstance(other, CompactGenome):
            return (self.mutations, self.reference) == (
                other.mutations,
                other.reference,
            )
        else:
            raise NotImplementedError

    def __le__(self, other: object) -> bool:
        if isinstance(other, CompactGenome):
            return sorted(self.mutations.items()) <= sorted(other.mutations.items())
        else:
            raise NotImplementedError

    def __lt__(self, other: object) -> bool:
        if isinstance(other, CompactGenome):
            return sorted(self.mutations.items()) < sorted(other.mutations.items())
        else:
            raise NotImplementedError

    def __gt__(self, other: object) -> bool:
        return not self.__le__(other)

    def __ge__(self, other: object) -> bool:
        return not self.__lt__(other)

    def __repr__(self):
        return (
            f"CompactGenome({self.mutations},"
            f" <reference sequence str with id:{id(self.reference)}>)"
        )

    def __str__(self):
        return f"CompactGenome[{', '.join(self.mutations_as_strings())}]"

    def get_site(self, site):
        """Get the base at the provided (one-based) site index."""
        mut = self.mutations.get(site)
        if mut is None:
            return self.reference[site - 1]
        else:
            return mut[-1]

    def mutations_as_strings(self):
        """Return mutations as a tuple of strings of the format '<reference
        base><index><new base>', sorted by index."""
        return tuple(
            (startbase + str(index) + endbase)
            for index, (startbase, endbase) in sorted(
                self.mutations.items(), key=lambda t: t[0]
            )
        )

    def mutate(self, mutstring: str, reverse: bool = False):
        """Apply a mutstring such as 'A110G' to this compact genome.

        In this example, A is the old base, G is the new base, and 110 is the 1-based
        index of the mutation in the sequence. Returns the new
        CompactGenome, and prints a warning if the old base doesn't
        match the recorded old base in this compact genome.

        Args:
            mutstring: The mutation to apply
            reverse: Apply the mutation in reverse, such as when the provided mutation
                describes how to achieve this CompactGenome from the desired CompactGenome.
        Returns:
            The new CompactGenome
        """
        oldbase = mutstring[0]
        newbase = mutstring[-1]
        if reverse:
            oldbase, newbase = newbase, oldbase
        idx = int(mutstring[1:-1])
        ref_base = self.reference[idx - 1]
        idx_present = idx in self.mutations
        if idx_present:
            old_recorded_base = self.mutations[idx][1]
        else:
            old_recorded_base = ref_base

        if oldbase != old_recorded_base:
            warn("recorded old base in sequence doesn't match old base")
        if ref_base == newbase:
            if idx_present:
                return CompactGenome(self.mutations.delete(idx), self.reference)
            else:
                return self
        return CompactGenome(
            self.mutations.set(idx, (ref_base, newbase)), self.reference
        )

    def apply_muts_raw(self, muts: Sequence[tuple]):
        """Apply the mutations from the sequence of tuples ``muts``.

        Each tuple should contain (one-based site, from_base, to_base)
        """
        # Is this tested? What does raw mean? Are there checks? What's mutate
        # for?
        res = dict(self.mutations)
        for site, from_base, to_base in muts:
            ref = self.reference[site - 1]
            if ref != to_base:
                res[site] = (ref, to_base)
            else:
                res.pop(site, None)
        return CompactGenome(res, self.reference)

    def apply_muts(self, muts: Sequence[str], reverse: bool = False, debug=False):
        """Apply a sequence of mutstrings like 'A110G' to this compact genome.

        In this example, A is the old base, G is the new base, and 110 is the 1-based
        index of the mutation in the sequence. Returns the new
        CompactGenome, and prints a warning if the old base doesn't
        match the recorded old base in this compact genome.

        Args:
            muts: The mutations to apply, in the order they should be applied
            reverse: Apply the mutations in reverse, such as when the provided mutations
                describe how to achieve this CompactGenome from the desired CompactGenome.
                If True, the mutations in `muts` will also be applied in reversed order.
            debug: If True, each mutation is applied individually by
                :meth:`CompactGenome.apply_mut` and the from base is checked against the
                current recorded base at each site.

        Returns:
            The new CompactGenome
        """
        newcg = self
        if reverse:
            mod_func = reversed

            def rearrange_func(tup):
                return tup[0], tup[2], tup[1]

        else:

            def mod_func(seq):
                yield from seq

            def rearrange_func(tup):
                return tup

        if debug:
            for mut in mod_func(muts):
                newcg = newcg.mutate(mut, reverse=reverse)
        else:
            newcg = self.apply_muts_raw(
                rearrange_func(unpack_mut_string(mut)) for mut in mod_func(muts)
            )

        return newcg

    def to_sequence(self):
        """Convert this CompactGenome to a full nucleotide sequence."""
        newseq = []
        newseq = list(self.reference)
        for idx, (ref_base, newbase) in self.mutations.items():
            if ref_base != newseq[idx - 1]:
                warn(
                    "CompactGenome.to_sequence warning: reference base doesn't match cg reference base"
                )
            newseq[idx - 1] = newbase
        return "".join(newseq)

    def mask_sites(self, sites, one_based=True):
        """Remove any mutations on sites in `sites`, leaving the reference
        sequence unchanged.

        Args:
            sites: A collection of sites to be masked
            one_based: If True, the provided sites will be interpreted as one-based sites. Otherwise,
                they will be interpreted as 0-based sites.
        """
        sites = set(sites)
        if one_based:

            def site_translate(site):
                return site

        else:

            def site_translate(site):
                return site - 1

        return CompactGenome(
            {
                site: data
                for site, data in self.mutations.items()
                if site_translate(site) not in sites
            },
            self.reference,
        )

    def superset_sites(self, sites, new_reference, one_based=True):
        """Do the opposite of `subset_sites`, adjusting site indices from
        indices in a sequence of variant sites, to indices in a sequence
        containing all sites.

        Args:
            sites: A sorted list of sites in the new_reference sequence which are represented
                by sites in the current compact genome's reference sequence
            new_reference: A new reference sequence
            one_based: Whether the sites in `sites` are one-based
        """
        if one_based:
            adjust = 0
        else:
            adjust = 1
        return CompactGenome(
            {sites[site - 1] + adjust: mut for site, mut in self.mutations.items()},
            new_reference,
        )

    def subset_sites(self, sites, one_based=True, new_reference=None):
        """Remove all but those sites in ``sites``, and adjust the reference
        sequence.

        Args:
            sites: A collection of sites to be kept
            one_based: If True, the provided sites will be interpreted as one-based sites. Otherwise,
                they will be interpreted as 0-based sites.
            new_reference: If provided, this new reference sequence will be used instead of
                computing the new reference sequence directly.
        """
        if one_based:
            adjust = 0
        else:
            adjust = 1
        site_map = {
            old_site + adjust: new_site
            for new_site, old_site in enumerate(sorted(sites), start=1)
        }
        result = {
            site_map[old_site]: state
            for old_site, state in self.mutations.items()
            if old_site - adjust in sites
        }
        if len(result) != len(self.mutations):
            warn("Sites where cg differed from reference removed.")

        if new_reference is None:
            new_reference = "".join(
                self.reference[site + adjust - 1] for site in sorted(sites)
            )

        return CompactGenome(result, new_reference)

    def remove_sites(self, sites, one_based=True, new_reference=None):
        """Remove all sites in ``sites``, and adjust the reference sequence.

        Args:
            sites: A collection of sites to be removed
            one_based: If True, the provided sites will be interpreted as one-based sites. Otherwise,
                they will be interpreted as 0-based sites.
            new_reference: If provided, this new reference sequence will be used instead of
                computing the new reference sequence directly.
        """
        if one_based:
            site_adjust = 0
        else:
            site_adjust = 1

        if new_reference is None:
            new_reference = "".join(
                base
                for site, base in enumerate(self.reference, start=1)
                if (site - site_adjust) not in sites
            )

        return CompactGenome(
            {
                mod_site: self.mutations[site]
                for mod_site, site in _iter_adjusted_sites(
                    self.mutations.keys(), sites, site_adjust
                )
            },
            new_reference,
        )


def unpack_mut_string(mut: str):
    """Returns (one-based site, from_base, to_base)"""
    return int(mut[1:-1]), mut[0], mut[-1]


def _iter_adjusted_sites(recorded_sites, removed_sites, site_adjust):
    """Adjusts recorded_sites if removed_sites are removed.

    For each one, returns a pair (modified site, unmodified site).
    site_adjust is the amount by which removed_sites base index is
    smaller than recorded_sites' base index.
    """
    all_sites = {site: False for site in recorded_sites}
    all_sites.update({site + site_adjust: True for site in removed_sites})
    shift = 0
    for site, removed in sorted(all_sites.items()):
        if removed:
            shift += 1
        else:
            yield site - shift, site


def compact_genome_from_sequence(sequence: str, reference: str):
    """Create a CompactGenome from a sequence and a reference sequence.

    Args:
        sequence: the sequence to be represented by a CompactGenome
        reference: the reference sequence for the CompactGenome
    """
    cg = {
        zero_idx + 1: (old_base, new_base)
        for zero_idx, (old_base, new_base) in enumerate(zip(reference, sequence))
        if old_base != new_base
    }
    return CompactGenome(cg, reference)


def cg_diff(parent_cg: CompactGenome, child_cg: CompactGenome):
    """Yields mutations in the format (parent_nuc, child_nuc, one-based
    sequence_index) distinguishing two compact genomes, such that applying the
    resulting mutations to `parent_cg` would yield `child_cg`"""
    keys = set(parent_cg.mutations.keys()) | set(child_cg.mutations.keys())
    for key in keys:
        parent_base = parent_cg.get_site(key)
        child_base = child_cg.get_site(key)
        if parent_base != child_base:
            yield (parent_base, child_base, key)


def ambiguous_cg_diff(
    parent_cg: CompactGenome,
    child_cg: CompactGenome,
    transition_model=default_nt_transitions,
    randomize=False,
):
    """Yields a minimal collection of mutations in the format (parent_nuc,
    child_nuc, one-based sequence_index) distinguishing two compact genomes,
    such that applying the resulting mutations to `parent_cg` would yield a
    compact genome compatible with the possibly ambiguous `child_cg`.

    If randomize is True, mutations will be randomized when there are
    multiple possible min-weight choices.
    """
    for parent_base, child_base, key in cg_diff(parent_cg, child_cg):
        nbase, _ = transition_model.min_character_mutation(
            parent_base,
            child_base,
            site=key,
            randomize=randomize,
        )
        if parent_base != nbase:
            yield (parent_base, nbase, key)


def reconcile_cgs(
    cg_list, check_references=True, ambiguitymap=standard_nt_ambiguity_map
):
    """Returns a compact genome containing ambiguous bases, representing the
    least ambiguous sequence of which all provided cgs in `cg_list` are
    resolutions. Also returns a flag indicating whether the resulting CG
    contains ambiguities.

    If `check_references` is False, reference sequences will be assumed equal.
    """
    ambiguous_flag = False
    if len(cg_list) == 1:
        return (cg_list[0], ambiguous_flag)

    cg_list = list(cg_list)
    model_cg = cg_list[0]
    reference = model_cg.reference
    if check_references:
        assert all(cg.reference == reference for cg in cg_list)

    # This can almost certainly be made more efficient
    difference_sites = dict()
    for ocg in cg_list[1:]:
        diff = list(cg_diff(model_cg, ocg))
        if len(diff) > 0:
            ambiguous_flag = True
        for parent_nuc, child_nuc, one_idx in diff:
            if one_idx not in difference_sites:
                difference_sites[one_idx] = {parent_nuc, child_nuc}
            else:
                # Parent nuc must already be added!
                difference_sites[one_idx].add(child_nuc)

    def process_charset(charset):
        diff = charset - ambiguitymap.bases
        if len(diff) > 0:
            charset.intersection_update(ambiguitymap.bases)
            for char in diff:
                charset.update(ambiguitymap[char])
        return frozenset(charset)

    mutstring_list = [
        model_cg.get_site(idx)
        + str(idx)
        + ambiguitymap.reversed[process_charset(charset)]
        for idx, charset in difference_sites.items()
    ]

    return (model_cg.apply_muts(mutstring_list), ambiguous_flag)


def read_alignment(alignment_file, reference_sequence: CharacterSequence = None):
    """Read a fasta or vcf alignment and return a dictionary mapping sequence
    ID strings to CompactGenomes.

    Args:
        alignment_file: A file containing a fasta or vcf alignment. File format
            is determined by extension. `.fa`, `.fasta`, or `.vcf` are expected.
        reference_sequence: If a fasta file is provided, the first sequence in that
            file will be used as the compact genome reference sequence, unless one
            is explicitly provided to this keyword argument.
    """
    extension = alignment_file.split(".")[-1].lower()

    if extension in ("fa", "fasta"):
        fasta_gen = read_fasta(alignment_file)
        cg_dict = {}
        if reference_sequence is None:
            # Now we just have to add an empty CG to the alignment
            refseq_id, reference_sequence = next(fasta_gen)
            cg_dict[refseq_id] = CompactGenome({}, reference_sequence)
        cg_dict.update(
            (seqid, compact_genome_from_sequence(seq, reference_sequence))
            for seqid, seq in fasta_gen
        )
    elif extension in ("vcf",):
        raise NotImplementedError
    else:
        raise ValueError(
            f"Unrecognized extension '.{extension}'. Provide a .fa, .fasta, or .vcf file."
        )
    return cg_dict



import historydag as hdag
from warnings import warn
import dendropy
import xml.etree.ElementTree as ET
import historydag.parsimony_utils as parsimony_utils


def dag_from_beast_trees(
    beast_xml_file,
    beast_output_file,
    reference_sequence=None,
    topologies_only=False,
    mask_ambiguous_sites=True,
    remove_ambiguous_sites=False,
    use_original_leaves=True,
    include_sequence_names_in_labels=False,
    transition_model=parsimony_utils.default_nt_transitions,
):
    """A convenience method to build a dag out of the output from
    :meth:`load_beast_trees`.

    Args:
        beast_xml_file: ''
        beast_output_file: ''
        reference_sequence: Provide a reference sequence, if desired. Otherwise, an arbitrary
            observed sequence will be used.
        topologies_only: If True, no internal sequences will be recovered from the beast output.
            In this case, leaf compact genomes will contain observed (possibly ambiguous)
            sequences regardless of the value of `use_original_leaves`.
        mask_ambiguous_sites: ''
        remove_ambiguous_sites: ''
        use_original_leaves: Use the original observed sequences for leaf node labels, instead
            of thos derived from simulated mutations.
        include_sequence_names_in_labels: If True, augment leaf node labels with a ``name`` attribute
            containing the name of the corresponding sequence. Useful for distinguishing leaves when
            observed sequences are not unique.
    """
    dp_trees = load_beast_trees(
        beast_xml_file,
        beast_output_file,
        topologies_only=topologies_only,
        reference_sequence=reference_sequence,
        mask_ambiguous_sites=mask_ambiguous_sites,
        remove_ambiguous_sites=remove_ambiguous_sites,
        transition_model=transition_model,
    )[0]

    if topologies_only:

        def cg_func(node):
            if node.is_leaf():
                return node.observed_cg
            else:
                return None

    elif use_original_leaves:

        def cg_func(node):
            if node.is_leaf():
                return node.observed_cg
            else:
                return node.cg

    else:

        def cg_func(node):
            return node.cg

    label_functions = {"compact_genome": cg_func}

    if include_sequence_names_in_labels:
        label_functions["name"] = lambda n: (
            n.taxon.label if n.is_leaf() else "internal"
        )

    dag = hdag.history_dag_from_trees(
        (tree.seed_node for tree in dp_trees),
        [],
        label_functions=label_functions,
        attr_func=lambda n: {"name": (n.taxon.label if n.is_leaf() else "internal")},
        child_node_func=dendropy.Node.child_nodes,
        leaf_node_func=dendropy.Node.leaf_iter,
    )
    if topologies_only:
        return dag
    else:
        return hdag.mutation_annotated_dag.AmbiguousLeafCGHistoryDag.from_history_dag(
            dag
        )


def load_beast_trees(
    beast_xml_file,
    beast_output_file,
    topologies_only=False,
    reference_sequence=None,
    mask_ambiguous_sites=True,
    remove_ambiguous_sites=False,
    transition_model=parsimony_utils.default_nt_transitions,
):
    """Load trees from BEAST output.

    Loads trees from BEAST output, in which each node has a `history_all` attribute
    containing the mutations inferred along that node's parent branch.

    Args:
        beast_xml_file: The xml input file to BEAST
        beast_output_file: The .trees output file from BEAST
        topologies_only: If True, no ancestral sequences are recovered from the `history_all`
            node attribute. This makes it possible to load trees which don't have that attribute.
        reference_sequence: If provided, a reference sequence which will be used for all
            compact genomes. By default, uses the ancestral sequence of the first tree, or if
            ``topologies_only`` is True, an arbitrary observed sequence.
        mask_ambiguous_sites: If True, ignore mutations for all sites whose observed set
            of characters is a subset of {N, -, ?} (recommended).
        remove_ambiguous_sites: If True, acts like ``mask_ambiguous_sites=True``, except
            the sites in question are actually removed from the sequence, rather than masked.

    Returns:
        A generator yielding :class:`dendropy.Tree`s output by BEAST, and a set of 0-based sites
        which are removed from sequences. If remove_ambiguous_sites is False, this set contains only
        sites ignored by BEAST. Otherwise, it also contains additional sites removed.
        Each tree has:
        * ancestral sequence attribute on each tree object, containing the complete reference
            for that tree
        * cg attribute on all nodes, containing a compact genome relative to the reference
            sequence
        * observed_cg attribute on leaf nodes, containing a compact genome describing the original
            observed sequence, with ambiguities, but with sites ignored by BEAST removed.
        * mut attribute on all nodes containing a list of mutations on parent branch, in
            order of occurrence
    """
    fasta, all_removed_sites = fasta_from_beast_file(
        beast_xml_file, remove_ignored_sites=True
    )

    all_removed_sites = set(all_removed_sites)
    # dendropy doesn't parse nested lists correctly in metadata, so we load the
    # trees with raw comment strings using `extract_comment_metadata`
    dp_trees = dendropy.TreeList.get(
        path=beast_output_file,
        schema="nexus",
        extract_comment_metadata=False,
        preserve_underscores=True,
    )

    def result_generator():
        # Get process_first, which recovers the tree ancestral sequence,
        # if necessary, and returns the correct reference sequence.
        if not topologies_only:
            if reference_sequence is None:

                def process_first(tree):
                    ref = _recover_reference(
                        tree, fasta, transition_model.ambiguity_map
                    )
                    return ref

            else:

                def process_first(tree):
                    _recover_reference(tree, fasta, transition_model.ambiguity_map)
                    return reference_sequence

        else:
            if reference_sequence is None:
                ref = next(iter(fasta.values()))

            def process_first(tree):
                return ref

        # Begin processing first tree and get reference sequence
        ref = process_first(dp_trees[0])

        if mask_ambiguous_sites or remove_ambiguous_sites:
            extra_masked_sites = {
                i
                for i in range(len(next(iter(fasta.values()))))
                if len(
                    {seq[i] for seq in fasta.values()}
                    - transition_model.ambiguity_map.uninformative_chars
                )
                == 0
            }
            if remove_ambiguous_sites:
                all_removed_sites.update(extra_masked_sites)
                new_reference = mask_sequence(ref, extra_masked_sites)

                def cg_transform(cg):
                    return cg.remove_sites(
                        extra_masked_sites, one_based=False, new_reference=new_reference
                    )

            elif mask_ambiguous_sites:

                def cg_transform(cg):
                    return cg.mask_sites(extra_masked_sites, one_based=False)

        else:

            def cg_transform(cg):
                return cg

        def get_observed_cgs(tree):
            for node in tree.leaf_node_iter():
                node.observed_cg = cg_transform(
                    hdag.compact_genome.compact_genome_from_sequence(
                        fasta[node.taxon.label], ref
                    )
                )

        if topologies_only:

            def process_second(tree):
                get_observed_cgs(tree)

        else:

            def process_second(tree):
                get_observed_cgs(tree)
                _recover_cgs(tree, ref, fasta, cg_transform, transition_model)

        # finish processing first tree:
        process_second(dp_trees[0])
        yield dp_trees[0]

        # process the rest:
        for tree in dp_trees[1:]:
            process_first(tree)
            process_second(tree)
            yield tree

    return result_generator(), all_removed_sites


def _recover_cgs(tree, reference_sequence, fasta, cg_transform, transition_model):
    ancestral_cg = hdag.compact_genome.compact_genome_from_sequence(
        tree.ancestral_sequence, reference_sequence
    )

    unmodified_node_cgs = {None: ancestral_cg}
    for node in tree.preorder_node_iter():
        parent_cg = unmodified_node_cgs[node.parent_node]
        this_cg = parent_cg.apply_muts_raw(node.muts)
        unmodified_node_cgs[node] = this_cg
        node.cg = cg_transform(this_cg)


def _comment_parser(node_comments):
    if len(node_comments) == 0:
        yield from ()
        return
    elif len(node_comments) == 1:
        comment_string = node_comments[0]
    else:
        for comment in node_comments:
            if "history_all" in comment:
                comment_string = comment
                break
        else:
            raise ValueError(
                "history_all node attribute not found:" + str(node_comments)
            )
    if "history_all=" not in comment_string:
        yield from ()
        return
    else:
        mutations_string = comment_string.split("history_all=")[-1]
        stripped_mutations_list = mutations_string[2:-2]
        if stripped_mutations_list:
            mutations_list = stripped_mutations_list.split("},{")
            for mut in mutations_list:
                try:
                    idx_str, _, from_base, to_base = mut.split(",")
                except ValueError:
                    raise ValueError("comment_parser failed on: " + str(node_comments))
                yield (int(idx_str), from_base, to_base)
        else:
            yield from ()
            return


def _recover_reference(tree, fasta, ambiguity_map):
    def get_least_ambiguous_base(idx, fasta):
        # looks at bases in fasta entries at (0-based) idx,
        # returns the first which is a concrete base, or the
        # intersection of all ambiguous bases
        seen = set()
        for val in fasta.values():
            base = val[idx]
            if base in ambiguity_map.bases:
                return base
            else:
                seen.add(base)
        intersection = set(ambiguity_map.bases)
        for code in seen:
            intersection.intersection_update(ambiguity_map[code])
        try:
            return ambiguity_map.reversed[frozenset(intersection)]
        except KeyError:
            warn(
                f"No ambiguity code found for possible reference bases {intersection} at idx {idx}"
            )
            return next(iter(intersection))

    ref_node = next(tree.postorder_node_iter())
    ref_sequence = list(fasta[ref_node.taxon.label])

    curr_sequence = [None] * len(ref_sequence)
    unknown_site_count = len(ref_sequence)
    for node in tree.preorder_node_iter():
        node.muts = list(_comment_parser(node.comments))
        if unknown_site_count > 0:
            for site, upbase, _ in node.muts:
                idx = site - 1
                if curr_sequence[idx] is None:
                    curr_sequence[idx] = upbase
                    unknown_site_count -= 1

    for idx, base in enumerate(curr_sequence):
        if base is None:
            curr_sequence[idx] = get_least_ambiguous_base(idx, fasta)

    tree.ancestral_sequence = "".join(curr_sequence)

    return tree.ancestral_sequence


def fasta_from_beast_file(filepath, remove_ignored_sites=True):
    """Produces an alignment dictionary from a BEAST xml file.

    Args:
        filepath: path to the BEAST xml file, containing an `alignment` block
        remove_ignored_sites: remove sites which are 'N' or '?' in all samples

    Returns:
        The resulting alignment dictionary, containing sequences keyed by names,
        and a tuple containing masked sites (this is empty if ``remove_ignored_sites``
        is False). Site indices are 0-based.
    """
    _etree = ET.parse(filepath)
    _alignment = _etree.getroot().find("alignment")
    unmasked_fasta = {
        a[0].attrib["idref"].strip(): a[0].tail.strip() for a in _alignment
    }
    masked_sites = {
        i
        for i in range(len(next(iter(unmasked_fasta.values()))))
        if len({seq[i] for seq in unmasked_fasta.values()} - {"N", "?"}) == 0
    }

    if remove_ignored_sites:
        return (
            {
                key: mask_sequence(val, masked_sites)
                for key, val in unmasked_fasta.items()
            },
            tuple(masked_sites),
        )
    else:
        return (unmasked_fasta, tuple())


def mask_sequence(unmasked, masked_sites):
    """Remove the 0-based indices in ``masked_sites`` from the sequence
    ``unmasked``, and return the resulting sequence."""
    return "".join(char for i, char in enumerate(unmasked) if i not in masked_sites)



from historydag.utils import (
    hist,
    AddFuncDict,
    prod,
)
from collections import Counter
from historydag.parsimony_utils import (
    hamming_distance_countfuncs,
    default_nt_transitions,
)

hamming_distance = default_nt_transitions.weighted_hamming_distance


def test_hamming_distance():
    assert hamming_distance("AGC", "AGC") == 0
    assert hamming_distance("ACC", "AGC") == 1
    try:
        hamming_distance("AG", "GCC")
    except ValueError:
        return
    raise ValueError(
        "hamming distance allowed comparison of sequences with different lengths."
    )


def test_hist():
    hist(Counter([1, 2, 3, 3, 3, 3, 4]))


def test_prod():
    assert prod([]) == 1


def test_dp_templatefuncs():
    templatedict = AddFuncDict(
        {
            "start_func": lambda n: 0,
            "edge_weight_func": lambda n, n1: int(n == n1),
            "accum_func": sum,
        },
        name="sum0",
    )
    templatedict1 = AddFuncDict(
        {
            "start_func": lambda n: 1,
            "edge_weight_func": lambda n, n1: int(n != n1),
            "accum_func": min,
        },
        name="min1",
    )
    template = templatedict + templatedict1

    startfunc, ewfunc, accumfunc = template.values()
    assert hamming_distance_countfuncs.names == ("HammingParsimony",)
    assert template.names == ("sum0", "min1")
    assert startfunc("doesnotmatter") == (0, 1)
    assert ewfunc("somenode", "someother") == (0, 1)
    assert accumfunc([(1, 2), (2, 3), (4, 5)]) == (7, 2)

    template1 = template + templatedict
    startfunc, ewfunc, accumfunc = template1.values()
    assert template1.names == ("sum0", "min1", "sum0")
    assert startfunc("doesnotmatter") == (0, 1, 0)
    assert ewfunc("somenode", "someother") == (0, 1, 0)
    assert accumfunc([(1, 2, 3), (2, 3, 4), (4, 5, 6)]) == (7, 2, 13)



"""Utility functions and classes for working with HistoryDag objects."""

import ete3
from math import log, exp, isfinite
from collections import Counter
from functools import wraps
import operator
from collections import UserDict
from decimal import Decimal
from warnings import warn
from typing import (
    List,
    Any,
    TypeVar,
    Callable,
    Union,
    Iterable,
    Generator,
    Tuple,
    NamedTuple,
    Optional,
)
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from historydag.dag import HistoryDagNode, HistoryDag

Weight = Any
Label = Union[NamedTuple, "UALabel"]
F = TypeVar("F", bound=Callable[..., Any])


class TaxaError(ValueError):
    pass


class UALabel(str):
    _fields: Tuple = tuple()

    def __new__(cls):
        return super(UALabel, cls).__new__(cls, "UA_Node")

    def __eq__(self, other):
        return isinstance(other, UALabel)

    def __hash__(self):
        return hash("UA_Node")

    def __iter__(self):
        raise RuntimeError("Attempted to iterate from dag root UALabel")

    def _asdict(self):
        raise RuntimeError("Attempted to iterate from dag root UALabel")


# ######## Decorators ########
def access_nodefield_default(fieldname: str, default: Any) -> Any:
    """A decorator for accessing label fields on a HistoryDagNode. Converts a
    function taking some label field's values as positional arguments, to a
    function taking HistoryDagNodes as positional arguments.

    Args:
        fieldname: The name of the label field whose value the function takes as arguments
        default: A value that should be returned if one of the arguments is the DAG UA node.

    For example, instead of
    `lambda n1, n2: default if n1.is_ua_node() or n2.is_ua_node() else func(n1.label.fieldname, n2.label.fieldname)`,
    this wrapper allows one to write `access_nodefield_default(fieldname, default)(func)`.
    """

    def decorator(func):
        @ignore_uanode(default)
        @access_field("label")
        @access_field(fieldname)
        @wraps(func)
        def wrapper(*args: Label, **kwargs: Any) -> Weight:
            return func(*args, **kwargs)

        return wrapper

    return decorator


def access_field(fieldname: str) -> Callable[[F], F]:
    """A decorator for conveniently accessing a field in a label.

    To be used instead of something like `lambda l1, l2:
    func(l1.fieldname, l2.fieldname)`. Instead just write
    `access_field(fieldname)(func)`. Supports arbitrarily many
    positional arguments, which are all expected to be labels
    (namedtuples) with field `fieldname`.
    """

    def decorator(func: F):
        @wraps(func)
        def wrapper(*args: Label, **kwargs: Any) -> Any:
            newargs = [getattr(label, fieldname) for label in args]
            return func(*newargs, **kwargs)

        return wrapper

    return decorator


def ignore_uanode(default: Any) -> Callable[[F], F]:
    """A decorator to return a default value if any argument is a UANode.

    For instance, to allow distance between two nodes to be zero if one
    is UANode
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args: "HistoryDagNode", **kwargs: Any):
            for node in args:
                if node.is_ua_node():
                    return default
            else:
                return func(*args, **kwargs)

        return wrapper

    return decorator


def explode_label(labelfield: str):
    """A decorator to make it easier to expand a Label by a certain field.

    Args:
        labelfield: the name of the field whose contents the wrapped function is expected to
            explode

    Returns:
        A decorator which converts a function which explodes a field value, into a function
        which explodes the whole label at that field.
    """

    def decorator(
        func: Callable[[Any], Iterable[Any]]
    ) -> Callable[[Label], Iterable[Label]]:
        @wraps(func)
        def wrapfunc(label, *args, **kwargs):
            Label = type(label)
            d = label._asdict()
            for newval in func(d[labelfield], *args, **kwargs):
                d[labelfield] = newval
                yield Label(**d)

        return wrapfunc

    return decorator


# ######## Distances and comparisons... ########


def cartesian_product(
    optionlist: List[Callable[[], Iterable]], accum=tuple()
) -> Generator[Tuple, None, None]:
    """The cartesian product of iterables in a list.

    Takes a list of functions which each return a fresh generator on
    options at that site, and returns a generator yielding tuples, which
    are elements of the cartesian product of the passed generators'
    contents.
    """
    if optionlist:
        for term in optionlist[0]():
            yield from cartesian_product(optionlist[1:], accum=(accum + (term,)))
    else:
        yield accum


def hist(c: Counter, samples: int = 1):
    """Pretty prints a counter Normalizing counts using the number of samples,
    passed as the argument `samples`."""
    ls = list(c.items())
    ls.sort()
    print("Weight\t| Frequency\n------------------")
    for weight, freq in ls:
        print(f"{weight}  \t| {freq if samples==1 else freq/samples}")


def is_collapsed(tree: ete3.TreeNode) -> bool:
    """Return whether the provided tree is collapsed.

    Collapsed means that any edge whose target is not a leaf node
    connects nodes with different sequences.
    """
    return not any(
        node.sequence == node.up.sequence and not node.is_leaf()
        for node in tree.iter_descendants()
    )


def collapse_adjacent_sequences(tree: ete3.TreeNode) -> ete3.TreeNode:
    """Collapse nonleaf nodes that have the same sequence."""
    # Need to keep doing this until the tree fully collapsed. See gctree for this!
    tree = tree.copy()
    to_delete = []
    for node in tree.get_descendants():
        # This must stay invariably hamming distance, since it's measuring equality of strings
        if not node.is_leaf() and node.up.sequence == node.sequence:
            to_delete.append(node)
    for node in to_delete:
        node.delete()
    return tree


class AddFuncDict(UserDict):
    """Container for function keyword arguments to
    :meth:`historydag.HistoryDag.weight_count`. This is primarily useful
    because it allows instances to be added. Passing the result to
    `weight_count` as keyword arguments counts the weights jointly. A
    :class:`historydag.utils.AddFuncDict` which may be passed as keyword
    arguments to :meth:`historydag.HistoryDag.weight_count`,
    :meth:`historydag.HistoryDag.trim_optimal_weight`, or
    :meth:`historydag.HistoryDag.optimal_weight_annotate` methods to trim or
    annotate a :meth:`historydag.HistoryDag` according to the weight that the
    contained functions implement.

    For example, `dag.weight_count(**(parsimony_utils.hamming_distance_countfuncs + make_newickcountfuncs()))`
    would return a Counter object in which the weights are tuples containing hamming parsimony and newickstrings.

    Args:
        initialdata: A dictionary containing functions keyed by "start_func", "edge_weight_func", and
            "accum_func". "start_func" specifies weight assigned to leaf HistoryDagNodes.
            "edge_weight_func" specifies weight assigned to an edge between two HistoryDagNodes, with the
            first argument the parent node, and the second argument the child node.
            "accum_func" specifies how to 'add' a list of weights. See :meth:`historydag.HistoryDag.weight_count`
            for more details.
        name: A string containing a name for the weight to be counted. If a tuple of weights will be returned,
            use ``names`` instead.
        names: A tuple of strings containing names for the weights to be counted, if a tuple of weights will
            be returned by passed functions. If only a single weight will be returned, use ``name`` instead.
    """

    requiredkeys = {"start_func", "edge_weight_func", "accum_func"}

    def __init__(self, initialdata, name: str = None, names: Tuple[str] = None):
        self.name: Optional[str]
        self.names: Tuple[str]
        if name is not None and names is not None:
            raise ValueError(
                "Pass a value to either keyword argument 'name' or 'names'."
            )
        elif name is None and names is None:
            self.name = "unknown weight"
            self.names = (self.name,)
        elif name is not None:
            self.name = name
            self.names = (self.name,)
        elif names is not None:
            if not isinstance(names, tuple):
                raise ValueError("``names`` keyword argument expects a tuple.")
            self.names = names
            self.name = None
        if not set(initialdata.keys()) == self.requiredkeys:
            raise ValueError(
                "Must provide functions named " + ", ".join(self.requiredkeys)
            )
        super().__init__(initialdata)

    def __add__(self, other) -> "AddFuncDict":
        fdict1 = self._convert_to_tupleargs()
        fdict2 = other._convert_to_tupleargs()
        n = len(fdict1.names)

        def newaccumfunc(weightlist):
            return fdict1["accum_func"](
                [weight[0:n] for weight in weightlist]
            ) + fdict2["accum_func"]([weight[n:] for weight in weightlist])

        def addfuncs(func1, func2):
            def newfunc(*args):
                return func1(*args) + func2(*args)

            return newfunc

        return AddFuncDict(
            {
                "start_func": addfuncs(fdict1["start_func"], fdict2["start_func"]),
                "edge_weight_func": addfuncs(
                    fdict1["edge_weight_func"], fdict2["edge_weight_func"]
                ),
                "accum_func": newaccumfunc,
            },
            names=fdict1.names + fdict2.names,
        )

    def __str__(self) -> str:
        return f"AddFuncDict[{', '.join(str(it) for it in self.names)}]"

    def _convert_to_tupleargs(self):
        if self.name is not None:

            def node_to_weight_decorator(func):
                @wraps(func)
                def wrapper(*args):
                    return (func(*args),)

                return wrapper

            def list_of_weights_to_weight_decorator(func):
                @wraps(func)
                def wrapper(weighttuplelist: List[Weight]):
                    return (func([wt[0] for wt in weighttuplelist]),)

                return wrapper

            return AddFuncDict(
                {
                    "start_func": node_to_weight_decorator(self["start_func"]),
                    "edge_weight_func": node_to_weight_decorator(
                        self["edge_weight_func"]
                    ),
                    "accum_func": list_of_weights_to_weight_decorator(
                        self["accum_func"]
                    ),
                },
                names=(self.name,),
            )
        else:
            return self

    def linear_combination(self, coeffs, significant_digits=8):
        """Convert an AddFuncDict implementing a tuple of weights to a linear
        combination of those weights.

        This only works when the weights computed by the AddFuncDict use plain
        `sum` as their accum_func.
        Otherwise, although the resulting AddFuncDict may be usable without errors,
        its behavior is undefined.

        Args:
            coeffs: The coefficients to be multiplied with each weight before summing.
            significant_digits: To combat floating point errors, only this many digits
                after the decimal will be significant in comparisons between weights.

        Returns:
            A new AddFuncDict object which computes the specified linear combination
            of weights.
        """
        n = len(self.names)
        if len(coeffs) != n:
            raise ValueError(
                f"Expected {n} ranking coefficients but received {len(coeffs)}."
            )
        if n == 1:
            raise ValueError(
                "linear_combination should only be called on AddFuncDict"
                " objects which compute more than one weight, e.g."
                " resulting from summing one or more AddFuncDicts."
            )

        def make_floatstate(val):
            return FloatState(round(val, significant_digits), state=val)

        def _lc(weight_tuple):
            return make_floatstate(sum(c * w for c, w in zip(coeffs, weight_tuple)))

        def accum_func(weights):
            return make_floatstate(sum(w.state for w in weights))

        start_func = self["start_func"]
        edge_func = self["edge_weight_func"]

        def new_start_func(n):
            return _lc(start_func(n))

        def new_edge_func(n1, n2):
            return _lc(edge_func(n1, n2))

        return AddFuncDict(
            {
                "start_func": new_start_func,
                "edge_weight_func": new_edge_func,
                "accum_func": accum_func,
            },
            name="("
            + " + ".join(
                str(c) + "(" + name + ")" for c, name in zip(coeffs, self.names)
            )
            + ")",
        )


class HistoryDagFilter:
    """Container for :class:`historydag.utils.AddFuncDict` and an optimality
    function `optimal_func`

    Args:
        weight_func: An :class:`AddFuncDict` object
        optimal_func: A function that specifies how to choose the optimal result from `weight_func`. e.g. `min` or `max`
    """

    def __init__(
        self,
        weight_funcs: AddFuncDict,
        optimal_func,
        ordering_name=None,
        eq_func=operator.eq,
    ):
        self.weight_funcs = weight_funcs
        self.optimal_func = optimal_func
        self.eq_func = eq_func
        end_idx = len(self.weight_funcs.names)
        if ordering_name is None:
            if optimal_func == min:
                self.ordering_names = (("minimum", end_idx),)
            elif optimal_func == max:
                self.ordering_names = (("maximum", end_idx),)
            else:
                self.ordering_names = (("optimal", end_idx),)
        else:
            self.ordering_names = ((ordering_name, end_idx),)

    def __str__(self) -> str:
        start_idx = 0
        descriptions = []
        for ordering_name, end_idx in self.ordering_names:
            these_names = self.weight_funcs.names[start_idx:end_idx]
            if len(these_names) > 1:
                descriptions.append(
                    f"{ordering_name} ({', '.join(str(it) for it in these_names)})"
                )
            else:
                descriptions.append(ordering_name + " " + these_names[0])
            start_idx = end_idx
        return "HistoryDagFilter[" + " then ".join(descriptions) + "]"

    def __getitem__(self, item):
        if item == "optimal_func":
            return self.optimal_func
        elif item == "eq_func":
            return self.eq_func
        else:
            return self.weight_funcs[item]

    # Or should it be &?
    def __add__(self, other):
        if not isinstance(other, HistoryDagFilter):
            raise TypeError(
                f"Can only add HistoryDagFilter to HistoryDagFilter, not f{type(other)}"
            )
        split_idx = len(self.weight_funcs.names)

        def new_optimal_func(weight_tuple_seq):
            weight_tuple_seq = tuple(weight_tuple_seq)
            first_optimal_val = self.optimal_func(
                t[:split_idx] for t in weight_tuple_seq
            )
            second_optimal_val = other.optimal_func(
                t[split_idx:]
                for t in weight_tuple_seq
                if self.eq_func(t[:split_idx], first_optimal_val)
            )
            return first_optimal_val + second_optimal_val

        if self.eq_func == operator.eq and other.eq_func == operator.eq:
            new_eq_func = operator.eq
        else:

            def new_eq_func(a, b):
                return self.eq_func(a[:split_idx], b[:split_idx]) and other.eq_func(
                    a[split_idx:], b[split_idx:]
                )

        ret = HistoryDagFilter(
            self.weight_funcs + other.weight_funcs,
            new_optimal_func,
            eq_func=new_eq_func,
        )
        ret.ordering_names = self.ordering_names + tuple(
            (name, idx + split_idx) for name, idx in other.ordering_names
        )
        return ret

    def keys(self):
        yield from self.weight_funcs.keys()
        yield from ("optimal_func", "eq_func")

    # def with_linear_combination_ordering(self, ranking_coeffs, eq_func=operator.eq):
    #     ranking_coeffs = tuple(ranking_coeffs)
    #     n = len(self.weight_funcs.names)
    #     if len(ranking_coeffs) != n:
    #         raise ValueError(f"Expected {n} ranking coefficients but received {len(ranking_coeffs)}.")

    #     def _lc(weight_tuple):
    #         return sum(c * w for c, w in zip(ranking_coeffs, weight_tuple))

    #     def new_optimal_func(weight_tuple_sequence):
    #         return min(weight_tuple_sequence, key=_lc)

    #     def new_eq_func(weight_tup1, weight_tup2):
    #         return eq_func(_lc(weight_tup1), _lc(weight_tup2))

    #     ret = HistoryDagFilter(self.weight_funcs, new_optimal_func, eq_func=new_eq_func)
    #     new_optimal_func_name = ("minimum ("
    #                              + '+'.join(str(c) + chr(97 + i) for i, c in enumerate(ranking_coeffs))
    #                              + ") for ("
    #                              + ','.join(chr(97 + i) for i in range(n))
    #                              + ") =")
    #     ret.ordering_names = ((new_optimal_func_name, n),)
    #     return ret


node_countfuncs = AddFuncDict(
    {
        "start_func": lambda n: 0,
        "edge_weight_func": lambda n1, n2: 1,
        "accum_func": sum,
    },
    name="NodeCount",
)
"""Provides functions to count the number of nodes in trees.

For use with :meth:`historydag.HistoryDag.weight_count`.
"""


def natural_edge_probability(parent, child):
    """Return the downward-conditional edge probability of the edge from parent
    to child.

    This is defined as 1/n, where n is the number of edges descending
    from the same child clade of ``parent`` as this edge.
    """
    if parent.is_ua_node():
        return 1 / len(list(parent.children()))
    else:
        eset = parent.clades[child.clade_union()]
        return 1 / len(eset.targets)


log_natural_probability_funcs = AddFuncDict(
    {
        "start_func": lambda n: 0,
        "edge_weight_func": lambda n1, n2: log(natural_edge_probability(n1, n2)),
        "accum_func": sum,
    },
    name="LogNaturalProbability",
)
"""Provides functions to count the probabilities of histories in a DAG,
according to the natural distribution induced by the DAG topology."""


def _process_rf_one_sided_coefficients(one_sided, one_sided_coefficients):
    rf_type_suffix = "distance"
    RFType = IntState

    if one_sided is None:
        # Only then will one_sided_coefficients be considered
        if one_sided_coefficients != (1, 1):
            rf_type_suffix = "nonstandard"
            # As long as both coefficients are integers, RF distances will
            # be integers. Otherwise, we need to allow floats by using
            # FloatState objects.
            if not all(isinstance(it, int) for it in one_sided_coefficients):
                RFType = FloatState
    elif one_sided.lower() == "left":
        one_sided_coefficients = (1, 0)
        rf_type_suffix = "left_difference"
    elif one_sided.lower() == "right":
        one_sided_coefficients = (0, 1)
        rf_type_suffix = "right_difference"
    else:
        raise ValueError(
            f"Argument `one_sided` must have value 'left', 'right', or None, not {one_sided}"
        )

    s, t = one_sided_coefficients
    return s, t, rf_type_suffix, RFType


def sum_rfdistance_funcs(
    reference_dag: "HistoryDag",
    rooted: bool = True,
    one_sided: str = None,
    one_sided_coefficients: Tuple[float, float] = (1, 1),
):
    """Provides functions to compute the sum over all histories in the provided
    reference DAG, of rooted RF distances to those histories.

    Args:
        reference_dag: The reference DAG. The sum will be computed over all RF
            distances to histories in this DAG
        rooted: If False, use edges' splits for RF distance computation. Otherwise, use
            the clade below each edge.
        one_sided: May be 'left', 'right', or None. 'left' means that we count
            splits (or clades, in the rooted case) which are in the reference trees but not
            in the DAG tree, especially useful if trees in the DAG might be resolutions of
            multifurcating trees in the reference DAG. 'right' means that we count splits or clades in
            the DAG tree which are not in the reference trees, useful if the reference trees
            are possibly resolutions of multifurcating trees in the DAG. If not None,
            one_sided_coefficients are ignored.
        one_sided_coefficients: coefficients for non-standard symmetric difference calculations
            (explained in notes below)

    The reference DAG must have the same taxa as all the trees in the DAG on which these count
    functions are used. If this is not true, methods using the keyword arguments produced by this
    function may fail silently, returning values which mean nothing.

    This function allows computation of sums of a Robinson-Foulds distance generalized by the
    coefficients ``(s, t)`` provided to the ``one_sided_coefficients`` argument (or implicitly
    set by the ``one_sided`` argument). Given a tree in the DAG with set of clades (or splits) A, and
    a tree in the reference DAG with set of clades B, this distance is given by:

    ``d_{s,t}(A, B) = s|B - A| + t|A - B|``

    Notice that when s and t are both 1, this is the symmetric difference of A and B, the standard RF
    distance.

    For each tree A in a DAG, the AddFuncDict returned by this function computes the sum of this distance
    over all trees B in the reference DAG.


    Note that when computing unrooted weights, the sums are over all rooted trees in the reference
    DAG, so a single unrooted tree contained twice in the reference DAG with different rootings
    will be counted twice.

    Weights are represented by an IntState object and are shifted by a constant K,
    which is the sum of number of clades in each tree in the DAG.
    """
    s, t, rf_type_suffix, RFType = _process_rf_one_sided_coefficients(
        one_sided, one_sided_coefficients
    )

    N = reference_dag.count_nodes(collapse=True, rooted=rooted)

    # K is the constant that the weights are shifted by
    K = s * sum(N.values())

    # We also scale num_trees by s...
    num_trees = t * reference_dag.count_histories()

    if rooted:

        def make_intstate(n):
            return RFType(n + K, state=n)

        def edge_func(n1, n2):
            clade = n2.clade_union()
            clade_count = N.get(clade, 0)
            weight = num_trees - ((s + t) * clade_count)
            return make_intstate(weight)

        kwargs = AddFuncDict(
            {
                "start_func": lambda n: make_intstate(0),
                "edge_weight_func": edge_func,
                "accum_func": lambda wlist: make_intstate(
                    sum(w.state for w in wlist)
                ),  # summation over edge weights
            },
            name="RF_rooted_sum_" + rf_type_suffix,
        )

    else:
        taxa = next(reference_dag.dagroot.children()).clade_union()
        n_taxa = len(taxa)

        def is_history_root(n):
            # TODO this is slow and dirty! Make more efficient
            return len(list(n.clade_union())) == n_taxa

        def split(node):
            cu = node.clade_union()
            return frozenset({cu, taxa - cu})

        # We accumulate tuples, where the first number contains the weight,
        # except any contribution of a split below a bifurcating root node
        # is contained in the second number. This way its contribution can be
        # added exactly once

        def make_intstate(tup):
            return RFType(tup[0] + tup[1] + K, state=tup)

        def summer(tupseq):
            tupseq = list(tupseq)
            a = 0
            for ia, _ in tupseq:
                a += ia
            # second value should only be counted once. Any nonzero
            # values of the second value will always be identical
            if len(tupseq) == 0:
                b = 0
            else:
                b = max(tupseq, key=lambda tup: abs(tup[1]))[1]
            return (a, b)

        def edge_func(n1, n2):
            spl = split(n2)
            spl_count = N.get(spl, 0)
            if n1.is_ua_node():
                return make_intstate((0, 0))
            else:
                val = num_trees - ((s + t) * spl_count)
                if len(n1.clades) == 2 and is_history_root(n1):
                    return make_intstate((0, val))
                else:
                    return make_intstate((val, 0))

        kwargs = AddFuncDict(
            {
                "start_func": lambda n: make_intstate((0, 0)),
                "edge_weight_func": edge_func,
                "accum_func": lambda wlist: make_intstate(
                    summer(w.state for w in wlist)
                ),  # summation over edge weights
            },
            name="RF_unrooted_sum_" + rf_type_suffix,
        )

    return kwargs


def make_rfdistance_countfuncs(
    ref_tree: "HistoryDag",
    rooted: bool = False,
    one_sided: str = None,
    one_sided_coefficients: Tuple[float, float] = (1, 1),
):
    """Provides functions to compute Robinson-Foulds (RF) distances of trees in
    a DAG, relative to a fixed reference tree.

    We use :meth:`ete3.TreeNode.robinson_foulds` as the reference implementation for
    unrooted RF distance.

    Rooted Robinson-Foulds is simply the cardinality of the symmetric difference of
    the clade sets of two trees, including the root clade.
    Since we include the root clade in this calculation, our rooted RF distance does
    not match the rooted :meth:`ete3.TreeNode.robinson_foulds` implementation.

    Args:
        ref_tree: A tree with respect to which Robinson-Foulds distance will be computed.
        rooted: If False, use edges' splits for RF distance computation. Otherwise, use
            the clade below each edge.
        one_sided: May be 'left', 'right', or None. 'left' means that we count
            splits (or clades, in the rooted case) which are in the reference tree but not
            in the DAG tree, especially useful if trees in the DAG might be resolutions of
            a multifurcating reference. 'right' means that we count splits or clades in
            the DAG tree which are not in the reference tree, useful if the reference tree
            is possibly a resolution of multifurcating trees in the DAG. If not None,
            one_sided_coefficients are ignored.
        one_sided_coefficients: coefficients for non-standard symmetric difference calculations
            (explained in notes below)

    The reference tree must have the same taxa as all the trees in the DAG.

    This calculation relies on the observation that the symmetric distance between
    the splits (or clades, in the rooted case) A in a tree in the DAG, and the splits
    (or clades) B in the reference tree, can be computed as:

    ``|B ^ A| = |B - A| + |A - B| = |B| - |A n B| + |A - B|``

    As long as tree edges are in bijection with splits, this can be computed without
    constructing the set A by considering each edge's split independently.

    In order to accommodate multiple edges with the same split in a tree with root
    bifurcation, we keep track of the contribution of such edges separately.

    One-sided RF distances are computed in this framework by introducing a pair of
    ``one_sided_coefficients`` ``(s, t)``, which affect how much weight is given to
    the right and left differences in the RF distance calculation:

    ``d_{s,t}(A, B) = s|B - A| + t|A - B| = s(|B| - |A n B|) + t|A - B|``

    When both ``s`` and ``t`` are 1, we get the standard RF distance.
    When ``s=1`` and ``t=0``, then we have a one-sided "left" RF difference, counting
    the number of splits in the reference tree which are not in each DAG tree. When
    ``one_sided`` is set to `left`, then these coefficients will be used, regardless of
    the values passed.
    When ``s=0`` and ``t=1``, then we have a one-sided "right" RF difference, counting
    the number of splits in each DAG tree which are not in the reference. When
    ``one_sided`` is set to `right`, these coefficients will be used, regardless of
    the values passed.

    The weight type is a tuple wrapped in an IntState object. The first tuple value `a` is the
    contribution of edges which are not part of a root bifurcation, where edges whose splits are in B
    contribute `-1`, and edges whose splits are not in B contribute `1`, and the second tuple
    value `b` is the contribution of the edges which are part of a root bifurcation. The value
    of the IntState is computed as `a + sign(b) + |B|`, which on the UA node of the hDAG gives RF distance.
    """

    s, t, rf_type_suffix, RFType = _process_rf_one_sided_coefficients(
        one_sided, one_sided_coefficients
    )

    taxa = frozenset(n.label for n in ref_tree.get_leaves())

    if not rooted:

        def split(node):
            cu = node.clade_union()
            return frozenset({cu, taxa - cu})

        ref_splits = frozenset(split(node) for node in ref_tree.preorder())
        # Remove above-root split, which doesn't map to any tree edge:
        ref_splits = ref_splits - {
            frozenset({taxa, frozenset()}),
        }
        shift = s * len(ref_splits)

        n_taxa = len(taxa)

        def is_history_root(n):
            # TODO this is slow and dirty! Make more efficient
            return len(list(n.clade_union())) == n_taxa

        def sign(n):
            # Should return the value of a single term corresponding
            # to the identical root splits below a bifurcating root
            return (-s) * (n < 0) + t * (n > 0)

        def summer(tupseq):
            a, b = 0, 0
            for ia, ib in tupseq:
                a += ia
                b += ib
            return (a, b)

        def make_intstate(tup):
            return RFType(tup[0] + shift + sign(tup[1]), state=tup)

        def edge_func(n1, n2):
            spl = split(n2)
            if n1.is_ua_node():
                return make_intstate((0, 0))
            if len(n1.clades) == 2 and is_history_root(n1):
                if spl in ref_splits:
                    return make_intstate((0, -1))
                else:
                    return make_intstate((0, 1))
            else:
                if spl in ref_splits:
                    return make_intstate((-s, 0))
                else:
                    return make_intstate((t, 0))

        kwargs = AddFuncDict(
            {
                "start_func": lambda n: make_intstate((0, 0)),
                "edge_weight_func": edge_func,
                "accum_func": lambda wlist: make_intstate(
                    summer(w.state for w in wlist)
                ),
            },
            name="RF_unrooted_distance_" + rf_type_suffix,
        )
    else:
        ref_cus = frozenset(
            node.clade_union() for node in ref_tree.preorder(skip_ua_node=True)
        )

        shift = s * len(ref_cus)

        def make_intstate(n):
            return RFType(n + shift, state=n)

        def edge_func(n1, n2):
            if n2.clade_union() in ref_cus:
                inval = 1
            else:
                inval = 0
            return make_intstate(t - (s + t) * inval)

        kwargs = AddFuncDict(
            {
                "start_func": lambda n: make_intstate(0),
                "edge_weight_func": edge_func,
                "accum_func": lambda wlist: make_intstate(sum(w.state for w in wlist)),
            },
            name="RF_rooted_" + rf_type_suffix,
        )

    return kwargs


def make_newickcountfuncs(
    name_func=lambda n: "unnamed",
    features=None,
    feature_funcs={},
    internal_labels=True,
    collapse_leaves=False,
):
    """Provides functions to count newick strings. For use with
    :meth:`historydag.HistoryDag.weight_count`.

    Arguments are the same as for
    :meth:`historydag.HistoryDag.to_newick`.
    """

    def _newicksum(newicks):
        # Filter out collapsed/deleted edges
        snewicks = sorted(newicks)
        if len(snewicks) == 2 and ";" in [newick[-1] for newick in snewicks if newick]:
            # Then we are adding an edge above a complete tree
            return "".join(
                sorted(snewicks, key=lambda n: ";" == n[-1] if n else False)
            )[:-1]
        else:
            # Then we're just accumulating options between clades
            return "(" + ",".join(snewicks) + ")"

    def _newickedgeweight(n1, n2):
        if collapse_leaves and n2.is_leaf() and n1.label == n2.label:
            return "COLLAPSED_LEAF;"
        elif (
            internal_labels
            or n2.is_leaf()
            or (collapse_leaves and frozenset({n2.label}) in n2.clades)
        ):
            return (
                n2._newick_label(
                    name_func=name_func, features=features, feature_funcs=feature_funcs
                )
                + ";"
            )
        else:
            return ";"

    return AddFuncDict(
        {
            "start_func": lambda n: "",
            "edge_weight_func": _newickedgeweight,
            "accum_func": _newicksum,
        },
        name="NewickString",
    )


def edge_difference_funcs(reference_dag: "HistoryDag", key=lambda n: n):
    """Provides functions to compute the number of edges in a history which do
    not appear in a reference HistoryDag.

    This is useful for taking history-wise intersections of DAGs, or counting
    the number of histories which would appear in such an intersection.

    Args:
        reference_dag: The reference DAG. These functions will count the
            number of edges in a history which do not appear in this DAG.

    Returns:
        :class:`utils.AddFuncDict` object for use with HistoryDag methods for
        trimming and weight counting/annotation.
    """
    edge_set = set(
        (key(n), key(c)) for n in reference_dag.preorder() for c in n.children()
    )

    def edge_weight_func(n1, n2):
        return int((key(n1), key(n2)) not in edge_set)

    return AddFuncDict(
        {
            "start_func": lambda n: 0,
            "edge_weight_func": edge_weight_func,
            "accum_func": sum,
        },
        name="EdgeDifference",
    )


def _history_method(method):
    """HistoryDagNode method decorator to ensure that the method is only run on
    history DAGs which are histories."""

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self.is_history():
            raise ValueError(
                "to_newick requires the history DAG to be a history. "
                "To extract newicks from a general DAG, see to_newicks"
            )
        else:
            return method(self, *args, **kwargs)

    return wrapper


def prod(ls: list):
    """Return product of elements of the input list.

    if passed list is empty, returns 1.
    """
    n = len(ls)
    if n > 0:
        accum = ls[0]
        if n > 1:
            for item in ls[1:]:
                accum *= item
    else:
        accum = 1
    return accum


def logsumexp(ls: List[float]):
    """A numerically stable implementation of logsumexp, similar to Scipy's."""
    if len(ls) == 1:
        return ls[0]
    max_log = max(ls)
    if not isfinite(max_log):
        max_log = 0

    exponentiated = [exp(a - max_log) for a in ls]
    shifted_log_sum = log(sum(exponentiated))
    return shifted_log_sum + max_log


# Unfortunately these can't be made with a class factory (just a bit too meta for Python)
# short of doing something awful like https://hg.python.org/cpython/file/b14308524cff/Lib/collections/__init__.py#l232
def _remstate(kwargs):
    if "state" not in kwargs:
        kwargs["state"] = None
    intkwargs = kwargs.copy()
    intkwargs.pop("state")
    return intkwargs


class IntState(int):
    """A subclass of int, with arbitrary, mutable state.

    State is provided to the constructor as the keyword argument
    ``state``. All other arguments will be passed to ``int``
    constructor. Instances should be functionally indistinguishable from
    ``int``.
    """

    def __new__(cls, *args, **kwargs):
        intkwargs = _remstate(kwargs)
        return super(IntState, cls).__new__(cls, *args, **intkwargs)

    def __init__(self, *args, **kwargs):
        self.state = kwargs["state"]

    def __copy__(self):
        return IntState(int(self), state=self.state)

    def __getstate__(self):
        return {"val": int(self), "state": self.state}

    def __setstate__(self, statedict):
        self.state = statedict["state"]


class FloatState(float):
    """A subclass of float, with arbitrary, mutable state.

    State is provided to the constructor as the keyword argument
    ``state``. All other arguments will be passed to ``float``
    constructor. Instances should be functionally indistinguishable from
    ``float``.
    """

    def __new__(cls, *args, **kwargs):
        intkwargs = _remstate(kwargs)
        return super(FloatState, cls).__new__(cls, *args, **intkwargs)

    def __init__(self, *args, **kwargs):
        self.state = kwargs["state"]

    def __copy__(self):
        return FloatState(float(self), state=self.state)

    def __getstate__(self):
        return {"val": float(self), "state": self.state}

    def __setstate__(self, statedict):
        self.state = statedict["state"]


class DecimalState(Decimal):
    """A subclass of ``decimal.Decimal``, with arbitrary, mutable state.

    State is provided to the constructor as the keyword argument
    ``state``. All other arguments will be passed to ``Decimal``
    constructor. Instances should be functionally indistinguishable from
    ``Decimal``.
    """

    def __new__(cls, *args, **kwargs):
        intkwargs = _remstate(kwargs)
        return super(DecimalState, cls).__new__(cls, *args, **intkwargs)

    def __init__(self, *args, **kwargs):
        self.state = kwargs["state"]

    def __copy__(self):
        return DecimalState(Decimal(self), state=self.state)

    def __getstate__(self):
        return {"val": Decimal(self), "state": self.state}

    def __setstate__(self, statedict):
        self.state = statedict["state"]


class StrState(str):
    """A subclass of string, with arbitrary, mutable state.

    State is provided to the constructor as the keyword argument
    ``state``. All other arguments will be passed to ``str``
    constructor. Instances should be functionally indistinguishable from
    ``str``.
    """

    def __new__(cls, *args, **kwargs):
        intkwargs = _remstate(kwargs)
        return super(StrState, cls).__new__(cls, *args, **intkwargs)

    def __init__(self, *args, **kwargs):
        self.state = kwargs["state"]

    def __copy__(self):
        return StrState(str(self), state=self.state)

    def __getstate__(self):
        return {"val": str(self), "state": self.state}

    def __setstate__(self, statedict):
        self.state = statedict["state"]


def count_labeled_binary_topologies(n):
    """Returns the number of binary topologies on n labeled leaves.

    In these topologies, left and right branches are not distinguished,
    and internal nodes are not ranked.
    """
    return prod(range(1, 2 * n - 2, 2))


def read_fasta(fastapath, sequence_type=str):
    """Load a fasta file as a generator which yields (sequence ID, sequence)
    pairs.

    The function ``sequence_type`` will be called on each sequence as it
    is read from the fasta file, and the resulting object will be yielded as the second
    item in each sequence record pair.
    """
    seqids = set()
    with open(fastapath, "r") as fh:
        seqid = None
        sequence = ""
        for line in fh:
            if line[0] == ">":
                if seqid is not None:
                    yield (seqid, sequence_type(sequence))
                    seqids.add(seqid)
                seqid = line[1:].strip()
                sequence = ""
                if seqid in seqids:
                    raise ValueError(
                        "Duplicate records with matching identifier in fasta file"
                    )
            else:
                if seqid is None and line.strip():
                    raise ValueError(
                        "First non-blank line in fasta does not contain identifier"
                    )
                else:
                    sequence += line.strip().upper()
        yield (seqid, sequence_type(sequence))


def load_fasta(fastapath, sequence_type=str):
    """Load a fasta file as a dictionary, with sequence ids as keys and
    sequences as values.

    The function ``sequence_type`` will be called on each sequence as it
    is read from the fasta file, and the returned objects will be the values
    in the resulting alignment dictionary.
    """
    return dict(read_fasta(fastapath, sequence_type=sequence_type))


def _deprecate_message(message):
    def _deprecate(func):
        @wraps(func)
        def deprecated(*args, **kwargs):
            warn(message)
            return func(*args, **kwargs)

        return deprecated

    return _deprecate



"""This module allows the loading and manipulation of Larch mutation annotated
DAG protobuf files.

The resulting history DAG contains labels with 'compact genomes', and a
'refseq' attribute describing a reference sequence and set of mutations
relative to the reference.
"""

from frozendict import frozendict
from historydag.dag import HistoryDag, HistoryDagNode, UANode, EdgeSet
import historydag.utils
from historydag.compact_genome import (
    CompactGenome,
    compact_genome_from_sequence,
    cg_diff,
    ambiguous_cg_diff,
    reconcile_cgs,
)
from historydag.parsimony_utils import (
    compact_genome_hamming_distance_countfuncs,
    leaf_ambiguous_compact_genome_hamming_distance_countfuncs,
    default_nt_transitions,
    standard_nt_ambiguity_map,
)
import historydag.dag_pb2 as dpb
import json
from math import log
from typing import NamedTuple, Callable


_pb_nuc_lookup = {0: "A", 1: "C", 2: "G", 3: "T"}
_pb_nuc_codes = {nuc: code for code, nuc in _pb_nuc_lookup.items()}


def _pb_mut_to_str(mut):
    """Unpack protobuf-encoded mutation into 1-indexed mutations string."""
    return (
        _pb_nuc_lookup[mut.par_nuc] + str(mut.position) + _pb_nuc_lookup[mut.mut_nuc[0]]
    )


class HDagJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, CompactGenome):
            return dict(obj.mutations)
        elif isinstance(obj, frozendict):
            return dict(obj)
        elif isinstance(obj, frozenset):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


class NodeIDHistoryDag(HistoryDag):
    """A HistoryDag subclass with node labels containing string ``node_id``
    fields.

    For leaf nodes this string is a unique leaf identifier, and for
    internal nodes this is a string representation of an integer node
    ID.
    """

    _required_label_fields = {"node_id": []}
    _default_args = frozendict(
        {
            "start_func": (lambda n: 0),
            "optimal_func": min,
            "accum_func": sum,
        }
    )


class CGHistoryDag(HistoryDag):
    """A HistoryDag subclass with node labels containing CompactGenome objects.

    The constructor for this class requires that each node label contain
    a 'compact_genome' field, which is expected to hold a
    :class:`compact_genome.CompactGenome` object.

    A HistoryDag containing 'sequence' node label fields may be
    automatically converted to this subclass by calling the class method
    :meth:`CGHistoryDag.from_dag`, providing the HistoryDag object to be
    converted, and the reference sequence to the keyword argument
    'reference'.

    This subclass provides specialized methods for interfacing with
    Larch's MADAG protobuf format
    """

    _required_label_fields = {
        "compact_genome": [
            (
                ("sequence",),
                lambda n, reference, **kwargs: compact_genome_from_sequence(
                    n.label.sequence, reference, **kwargs
                ),
            )
        ]
    }

    _default_args = frozendict(compact_genome_hamming_distance_countfuncs) | {
        "start_func": (lambda n: 0),
        "optimal_func": min,
    }
    # #### Overridden Methods ####

    def weight_counts_with_ambiguities(self, *args, **kwargs):
        raise NotImplementedError(
            "This method is only implemented for DAGs with node labels containing sequences."
        )

    def summary(self):
        HistoryDag.summary(self)
        min_pars, max_pars = self.weight_range_annotate(
            **compact_genome_hamming_distance_countfuncs
        )
        print(f"Parsimony score range {min_pars} to {max_pars}")

    def hamming_parsimony_count(self):
        """See :meth:`historydag.sequence_dag.SequenceHistoryDag.hamming_parsim
        ony_count`"""
        return self.weight_count(**compact_genome_hamming_distance_countfuncs)

    # #### CGHistoryDag-Specific Methods ####

    def _get_mut_func(self):
        refseq = self.get_reference_sequence()
        empty_cg = CompactGenome(dict(), refseq)

        def mut_func(pnode, cnode, **kwargs):
            if pnode.is_ua_node():
                parent_seq = empty_cg
            else:
                parent_seq = pnode.label.compact_genome
            return cg_diff(parent_seq, cnode.label.compact_genome)

        return mut_func

    def to_protobuf(
        self,
        leaf_data_func=None,
        randomize_leaf_muts=False,
        transition_model=default_nt_transitions,
    ):
        """Convert a DAG with compact genome data on each node, and unique leaf
        IDs on leaf nodes, to a MAD protobuf with mutation information on
        edges.

        Args:
            leaf_data_func: a function taking a DAG node and returning a string to store
                in the protobuf node_name field `condensed_leaves` of leaf nodes. On leaf
                nodes, this data is appended after the unique leaf ID.
            randomize_leaf_muts: When leaf node sequences contain ambiguities, if
                True the mutations on pendant edges will be randomized, when
                there are multiple choices.
            transition_model: A :meth:`historydag.parsimony_utils.TransitionModel` object,
                used to decide which bases to record on pendant edge mutations with
                ambiguous bases as targets.

        Note that internal node IDs will be reassigned, even if internal nodes have
        node IDs in their label data.
        """

        mut_func = self._get_mut_func()

        # Create unique leaf IDs if the node_id field isn't available
        if "node_id" in self.get_label_type()._fields:

            def get_leaf_id(node):
                return node.label.node_id

        else:
            leaf_id_map = {n: f"s{idx}" for idx, n in enumerate(self.get_leaves())}

            def get_leaf_id(node):
                return leaf_id_map[node]

        def key_func(cladeitem):
            clade, _ = cladeitem
            return sorted(
                sorted(idx for idx in label.compact_genome.mutations) for label in clade
            )

        node_dict = {}
        data = dpb.data()
        for idx, node in enumerate(self.postorder()):
            node_dict[node] = idx
            node_name = data.node_names.add()
            node_name.node_id = idx
            if node.is_leaf():
                node_name.condensed_leaves.append(get_leaf_id(node))
                if leaf_data_func is not None:
                    node_name.condensed_leaves.append(leaf_data_func(node))

        for node in self.postorder():
            for cladeidx, (clade, edgeset) in enumerate(
                sorted(node.clades.items(), key=key_func)
            ):
                for child in edgeset.targets:
                    edge = data.edges.add()
                    edge.parent_node = node_dict[node]
                    edge.parent_clade = cladeidx
                    edge.child_node = node_dict[child]
                    for par_nuc, child_nuc, idx in mut_func(
                        node,
                        child,
                        randomize=randomize_leaf_muts,
                        transition_model=transition_model,
                    ):
                        mut = edge.edge_mutations.add()
                        mut.position = idx
                        mut.par_nuc = _pb_nuc_codes[par_nuc.upper()]
                        mut.mut_nuc.append(_pb_nuc_codes[child_nuc.upper()])
        data.reference_seq = self.get_reference_sequence()
        data.reference_id = (
            self.attr["refseqid"] if "refseqid" in self.attr else "unknown_seqid"
        )
        return data

    def to_protobuf_file(
        self, filename, leaf_data_func=None, randomize_leaf_muts=False
    ):
        """Write this CGHistoryDag to a Mutation Annotated DAG protobuf for use
        with Larch."""
        data = self.to_protobuf(
            leaf_data_func=leaf_data_func, randomize_leaf_muts=randomize_leaf_muts
        )
        with open(filename, "wb") as fh:
            fh.write(data.SerializeToString())

    def flatten(self, sort_compact_genomes=False):
        """Return a dictionary containing four keys:

        * `refseq` is a list containing the reference sequence id, and the reference sequence
          (the implied sequence on the UA node)
        * `compact_genome_list` is a list of compact genomes, where each compact genome is a
          list of nested lists `[seq_idx, [old_base, new_base]]` where `seq_idx` is (1-indexed)
          nucleotide sequence site. If sort_compact_genomes is True, compact genomes and `compact_genome_list` are sorted.
        * `node_list` is a list of `[label_idx, clade_list]` pairs, where
            * `label_idx` is the index of the node's compact genome in `compact_genome_list`, and
            * `clade_list` is a list of lists of `compact_genome_list` indices, encoding sets of child clades.

        * `edge_list` is a list of triples `[parent_idx, child_idx, clade_idx]`, where
            * `parent_idx` is the index of the edge's parent node in `node_list`,
            * `child_idx` is the index of the edge's child node in `node_list`, and
            * `clade_idx` is the index of the clade in the parent node's `clade_list` from which this edge descends.
        """
        compact_genome_list = []
        node_list = []
        edge_list = []
        node_indices = {}
        cg_indices = {}

        def get_child_clades(node):
            return [
                frozenset(cg_indices[label] for label in clade) for clade in node.clades
            ]

        def get_compact_genome(node):
            if node.is_ua_node():
                return []
            else:
                ret = [
                    [idx, list(bases)]
                    for idx, bases in node.label.compact_genome.mutations.items()
                ]

            if sort_compact_genomes:
                ret.sort()
            return ret

        for node in self.postorder():
            node_cg = get_compact_genome(node)
            if node.label not in cg_indices:
                cg_indices[node.label] = len(compact_genome_list)
                compact_genome_list.append(node_cg)

        if sort_compact_genomes:
            cgindexlist = sorted(enumerate(compact_genome_list), key=lambda t: t[1])
            compact_genome_list = [cg for _, cg in cgindexlist]
            # the rearrangement is a bijection of indices
            indexmap = {
                oldidx: newidx for newidx, (oldidx, _) in enumerate(cgindexlist)
            }
            for key in cg_indices:
                cg_indices[key] = indexmap[cg_indices[key]]

        for node_idx, node in enumerate(self.postorder()):
            node_indices[id(node)] = node_idx
            node_list.append((cg_indices[node.label], get_child_clades(node)))
            for clade_idx, (clade, eset) in enumerate(node.clades.items()):
                for child in eset.targets:
                    edge_list.append((node_idx, node_indices[id(child)], clade_idx))

        if "refseq" in self.attr:
            refseqid = self.attr["refseq"]
        else:
            refseqid = "unknown_seqid"
        return {
            "refseq": (refseqid, self.get_reference_sequence()),
            "compact_genomes": compact_genome_list,
            "nodes": node_list,
            "edges": edge_list,
        }

    def test_equal(self, other):
        """Deprecated test for whether two history DAGs are equal.

        Compares sorted JSON representation. Only works when
        "compact_genome" is the only label field, on all nodes.
        """
        flatdag1 = self.flatten()
        flatdag2 = other.flatten()
        cg_list1 = flatdag1["compact_genomes"]
        cg_list2 = flatdag2["compact_genomes"]

        def get_edge_set(flatdag):
            edgelist = flatdag["edges"]
            nodelist = flatdag["nodes"]

            def convert_flatnode(flatnode):
                label_idx, clade_list = flatnode
                clades = frozenset(
                    frozenset(label_idx_list) for label_idx_list in clade_list
                )
                return (label_idx, clades)

            nodelist = [convert_flatnode(node) for node in nodelist]
            return frozenset(
                (nodelist[p_idx], nodelist[c_idx]) for p_idx, c_idx, _ in edgelist
            )

        return cg_list1 == cg_list2 and get_edge_set(flatdag1) == get_edge_set(flatdag2)

    def get_reference_sequence(self):
        """Return the reference sequence for this CGHistoryDag.

        This is the sequence with respect to which all node label
        CompactGenomes record mutations.
        """
        return next(self.preorder(skip_ua_node=True)).label.compact_genome.reference

    def _check_valid(self, *args, **kwargs):
        assert super()._check_valid(*args, **kwargs)
        reference = self.get_reference_sequence()
        for node in self.preorder(skip_ua_node=True):
            if node.label.compact_genome.reference != reference:
                raise ValueError(
                    "Multiple compact genome reference sequences found in node label CompactGenomes."
                )
        return True

    def to_json(self, sort_compact_genomes=False):
        """Write this history DAG to a JSON object."""
        return json.dumps(
            self.flatten(sort_compact_genomes=sort_compact_genomes), cls=HDagJSONEncoder
        )

    def to_json_file(self, filename, sort_compact_genomes=False):
        """Write this history DAG to a JSON file."""
        with open(filename, "w") as fh:
            fh.write(self.to_json(sort_compact_genomes=sort_compact_genomes))

    def adjusted_node_probabilities(
        self,
        log_probabilities=False,
        ua_node_val=None,
        adjust_func: Callable[[HistoryDagNode, HistoryDagNode], float] = None,
        **kwargs,
    ):
        """Compute the probability of each node in the DAG, adjusted based on
        the frequency of mutations that define each node.

        See :meth:`HistoryDag.node_probabilities` for argument
        descriptions.
        """
        if adjust_func is None:
            uncollapsed = False
            mut_freq = {}  # (parent_nuc, child_nuc, sequence_index) -> frequency
            edge_counts = self.count_edges()
            total_muts = 0
            for child in reversed(list(self.postorder())):
                if not child.is_root():
                    for parent in child.parents:
                        if parent.is_root() or child.is_leaf():
                            continue
                        muts = list(
                            cg_diff(
                                parent.label.compact_genome, child.label.compact_genome
                            )
                        )
                        if len(muts) == 0:
                            uncollapsed = True

                        for mut in muts:
                            if mut not in mut_freq:
                                mut_freq[mut] = 0
                            mut_freq[mut] += edge_counts[(parent, child)]
                            total_muts += edge_counts[(parent, child)]

            if uncollapsed:
                raise Warning("Support adjustment on uncollapsed DAG.")

            min_mut_freq = 1
            for mut in mut_freq.keys():
                mut_freq[mut] /= total_muts
                assert mut_freq[mut] <= 1 and mut_freq[mut] >= 1 / total_muts
                if mut_freq[mut] < min_mut_freq:
                    min_mut_freq = mut_freq[mut]

            # TODO: Inspect this further to gather stats about what type of mutations are most common
            # print(mut_freq)

            # Returns a value in [0, 1] that indicates the correct adjustment
            if log_probabilities:

                def adjust_func(parent, child, min_mut_freq=min_mut_freq, eps=1e-2):
                    if parent.is_root() or child.is_leaf():
                        return 0
                    else:
                        diff = [
                            mut
                            for mut in cg_diff(
                                parent.label.compact_genome, child.label.compact_genome
                            )
                        ]
                        if len(diff) == 0:
                            return log(eps * min_mut_freq)
                        else:
                            return log(
                                1
                                - historydag.utils.prod([mut_freq[mut] for mut in diff])
                            )

            else:

                def adjust_func(parent, child, min_mut_freq=min_mut_freq, eps=1e-2):
                    if parent.is_root() or child.is_leaf():
                        return 1
                    else:
                        diff = [
                            mut
                            for mut in cg_diff(
                                parent.label.compact_genome, child.label.compact_genome
                            )
                        ]
                        if len(diff) == 0:
                            return eps * min_mut_freq
                        return 1 - historydag.utils.prod(
                            [mut_freq[mut] for mut in diff]
                        )

        return self.node_probabilities(
            log_probabilities=log_probabilities,
            adjust_func=adjust_func,
            ua_node_val=ua_node_val,
            **kwargs,
        )


class AmbiguousLeafCGHistoryDag(CGHistoryDag):
    """A HistoryDag subclass with node labels containing compact genomes.

    The constructor for this class requires that each node label contain
    a 'compact_genome' field, which is expected to hold a
    :class:`compact_genome.CompactGenome` object, which is expected to
    hold an unambiguous sequence if the node is internal. The sequence
    may contain ambiguities if the node is a leaf.

    A HistoryDag containing 'sequence' node label fields may be
    automatically converted to this subclass by calling the class method
    :meth:`CGHistoryDag.from_dag`, providing the HistoryDag object to be
    converted, and the reference sequence to the keyword argument
    'reference'.
    """

    _default_args = frozendict(
        leaf_ambiguous_compact_genome_hamming_distance_countfuncs
    ) | {
        "start_func": (lambda n: 0),
        "optimal_func": min,
    }

    # #### Overridden Methods ####

    def _get_mut_func(self):
        refseq = self.get_reference_sequence()
        empty_cg = CompactGenome(dict(), refseq)

        def mut_func(
            pnode, cnode, randomize=False, transition_model=default_nt_transitions
        ):
            if pnode.is_ua_node():
                parent_seq = empty_cg
            else:
                parent_seq = pnode.label.compact_genome
            if cnode.is_leaf():
                # have to choose non-ambiguous mutations that minimize
                # edge weight.
                return ambiguous_cg_diff(
                    parent_seq,
                    cnode.label.compact_genome,
                    randomize=randomize,
                    transition_model=transition_model,
                )
            else:
                return cg_diff(parent_seq, cnode.label.compact_genome)

        return mut_func

    def hamming_parsimony_count(self):
        """See :meth:`historydag.sequence_dag.SequenceHistoryDag.hamming_parsim
        ony_count`"""
        return self.weight_count(
            **leaf_ambiguous_compact_genome_hamming_distance_countfuncs
        )

    def summary(self):
        HistoryDag.summary(self)
        min_pars, max_pars = self.weight_range_annotate(
            **leaf_ambiguous_compact_genome_hamming_distance_countfuncs
        )
        print(f"Parsimony score range {min_pars} to {max_pars}")

    # #### End Overridden Methods ####


def load_json_file(filename):
    """Load a Mutation Annotated DAG stored in a JSON file and return a
    CGHistoryDag."""
    with open(filename, "r") as fh:
        json_dict = json.load(fh)
    return unflatten(json_dict)


def unflatten(flat_dag):
    """Takes a dictionary like that returned by flatten, and returns a
    HistoryDag."""
    refseqid, reference = flat_dag["refseq"]
    compact_genome_list = [
        CompactGenome({idx: tuple(bases) for idx, bases in flat_cg}, reference)
        for flat_cg in flat_dag["compact_genomes"]
    ]
    node_list = flat_dag["nodes"]
    edge_list = flat_dag["edges"]
    Label = NamedTuple("Label", [("compact_genome", CompactGenome)])

    def unpack_cladelabellists(cladelabelsetlist):
        return [
            frozenset(Label(compact_genome_list[cg_idx]) for cg_idx in idx_clade)
            for idx_clade in cladelabelsetlist
        ]

    node_postorder = []
    # a list of (node, [(clade, eset), ...]) tuples
    for cg_idx, cladelabellists in node_list:
        clade_eset_list = [
            (clade, EdgeSet()) for clade in unpack_cladelabellists(cladelabellists)
        ]
        if len(clade_eset_list) == 1:
            # This must be the UA node
            label = historydag.utils.UALabel()
        else:
            label = Label(compact_genome_list[cg_idx])
        try:
            node = HistoryDagNode(label, dict(clade_eset_list), attr=None)
        except ValueError:
            node = UANode(clade_eset_list[0][1])
        node_postorder.append((node, clade_eset_list))

    # adjust UA node label
    node_postorder[-1][0].label = historydag.utils.UALabel()

    # Add edges
    for parent_idx, child_idx, clade_idx in edge_list:
        node_postorder[parent_idx][1][clade_idx][1].add_to_edgeset(
            node_postorder[child_idx][0]
        )

    # UA node is last in postorder
    dag = CGHistoryDag(node_postorder[-1][0])
    dag.attr["refseq"] = refseqid
    # This shouldn't be necessary, but appears to be
    dag.recompute_parents()
    return dag


def load_MAD_protobuf(
    pbdata,
    compact_genomes=False,
    node_ids=True,
    leaf_cgs={},
    ambiguity_map=standard_nt_ambiguity_map,
):
    """Convert a Larch MAD protobuf to a CGLeafIDHistoryDag with compact
    genomes in the `compact_genome` label attribute.

    Args:
        pbdata: loaded protobuf data object
        compact_genomes: If True, returns a CGHistoryDag or AmbiguousLeafCGHistoryDag
            object, with labels containing `node_id` and `compact_genome` fields.
            If no leaf sequence data is provided, leaf compact genomes will be
            inferred from pendant edge mutations, and will include ambiguities when
            mutations on two pendant edges pointing to the same leaf would otherwise
            contradict. `node_id` field on internal nodes will be None, unless
            `node_ids` argument is True. If False, this function will return a NodeIDHistoryDag.
        node_ids: If True, node IDs will be included on all nodes' labels. If False, internal
            nodes' `node_id` label fields will be `None`. Unique leaf sequence IDs are always
            included in the `node_id` label field of leaf nodes, to ensure that leaf node
            labels are unique.
        leaf_cgs: (not implemented) A dictionary keyed by unique string leaf IDs containing CompactGenomes.
            Use :meth:`compact_genome.read_alignment` to read an alignment from a file.
        ambiguity_map: A :meth:`historydag.parsimony_utils.AmbiguityMap` object
            to determine how conflicting pendant edge mutations are represented.

    Note that if leaf sequences in the original alignment do not contain ambiguities, it is not
    necessary to provide alignment data; leaf sequences can be completely inferred without it.
    """

    class PBDAG:
        """This class doesn't do much, just provides an interface to the
        protobuf data, as a DAG of integer node IDs, and provides methods to
        get history DAG node data for a given node ID."""

        def __init__(self, pbdata):
            self.pbdata = pbdata
            self.reference = pbdata.reference_seq
            parent_edges = {node.node_id: [] for node in pbdata.node_names}
            # a list of list of a node's child edges
            child_edges = {node.node_id: [] for node in pbdata.node_names}
            for edge in pbdata.edges:
                parent_edges[edge.child_node].append(edge)
                child_edges[edge.parent_node].append(edge)
            self.child_edges = child_edges
            self.parent_edges = parent_edges

            # now each node id is in parent_edges and child_edges as a key,
            # fix the UA node's compact genome (could be done in function but this
            # asserts only one node has no parent edges)
            (ua_node_id,) = [
                node_id for node_id, eset in parent_edges.items() if len(eset) == 0
            ]
            self.ua_node_id = ua_node_id

            def traverse_postorder(node_id):
                # order node_ids in postordering
                visited = set()

                def traverse(node_id):
                    visited.add(node_id)
                    child_ids = [edge.child_node for edge in child_edges[node_id]]
                    if len(child_ids) > 0:
                        for child_id in child_ids:
                            if child_id not in visited:
                                yield from traverse(child_id)
                    yield node_id

                yield from traverse(node_id)

            self.id_postorder = list(traverse_postorder(ua_node_id))
            self.id_reverse_postorder = list(reversed(self.id_postorder))
            self.node_id_to_cg = None

            if node_ids:

                def _id_func(nid):
                    if self.is_leaf(nid):
                        return pbdata.node_names[node_id].condensed_leaves[0]
                    else:
                        return str(nid)

            else:

                def _id_func(nid):
                    if self.is_leaf(nid):
                        return pbdata.node_names[node_id].condensed_leaves[0]
                    else:
                        return None

            if compact_genomes:
                self.return_type = CGHistoryDag
                self.label_fields = ("compact_genome", "node_id")
                self._label_funcs = (self.get_compact_genome, _id_func)
            else:
                self.label_fields = ("node_id",)
                self._label_funcs = (_id_func,)
                self.return_type = NodeIDHistoryDag

        def _build_compact_genomes(self):
            # These are built from ua node down, so must be built in
            # reverse postorder:
            # Also returns flag indicating if any compact genomes are ambiguous
            node_id_to_cg = {
                self.ua_node_id: CompactGenome(frozendict(), self.reference)
            }
            assert self.id_reverse_postorder[0] == self.ua_node_id
            ambiguous_flag = False

            def get_leaf_cg(node_id):
                leaf_id = pbdata.node_names[node_id].condensed_leaves[0]
                cg = leaf_cgs.get(leaf_id, None)
                if cg is not None:
                    # If we're provided a leaf sequence dictionary, just assume
                    # leaf sequences are ambiguous (otherwise it would be
                    # unnecessary, except for speedup). If this assumption is
                    # untrue, it can be fixed with HistoryDag subtype conversion
                    return cg, True
                else:
                    edges = self.parent_edges[node_id]
                    str_mutations = [
                        tuple(_pb_mut_to_str(mut) for mut in edge.edge_mutations)
                        for edge in edges
                    ]
                    return reconcile_cgs(
                        [
                            node_id_to_cg[edge.parent_node].apply_muts(muts)
                            for edge, muts in zip(edges, str_mutations)
                        ],
                        ambiguitymap=ambiguity_map,
                    )

            for node_id in self.id_reverse_postorder[1:]:
                if len(self.child_edges[node_id]) > 0:
                    edge = self.parent_edges[node_id][0]
                    parent_cg = node_id_to_cg[edge.parent_node]
                    str_mutations = tuple(
                        _pb_mut_to_str(mut) for mut in edge.edge_mutations
                    )
                    node_id_to_cg[node_id] = parent_cg.apply_muts(str_mutations)
                else:
                    # node_id belongs to leaf, must look at all parent edges to
                    # look for contradictions (implying ambiguities)
                    node_id_to_cg[node_id], ambig_flag = get_leaf_cg(node_id)
                    ambiguous_flag = ambiguous_flag or ambig_flag

            self.node_id_to_cg = node_id_to_cg
            if ambiguous_flag:
                self.return_type = AmbiguousLeafCGHistoryDag

        def get_compact_genome(self, node_id):
            if self.node_id_to_cg is None:
                self._build_compact_genomes()
            return self.node_id_to_cg[node_id]

        def get_label(self, node_id):
            return tuple([lfunc(node_id) for lfunc in self._label_funcs])

        def is_leaf(self, node_id):
            return len(self.child_edges[node_id]) == 0

        def get_children(self, node_id):
            return [edge.child_node for edge in self.child_edges[node_id]]

        def get_parents(self, node_id):
            return [edge.parent_node for edge in self.parent_edges[node_id]]

    pbdag = PBDAG(pbdata)

    node_to_node_d = dict()
    node_id_to_node = dict()
    Label = NamedTuple("Label", [(label, any) for label in pbdag.label_fields])  # type: ignore

    for node_id in pbdag.id_postorder:
        # These have all been created already
        children = [
            node_id_to_node[child_id] for child_id in pbdag.get_children(node_id)
        ]
        child_clades = frozenset({child.clade_union() for child in children})
        if node_id == pbdag.ua_node_id:
            this_node = UANode(EdgeSet())
        else:
            this_node = HistoryDagNode(
                Label(*pbdag.get_label(node_id)),
                {child_clade: EdgeSet() for child_clade in child_clades},
                {"node_id": node_id},
            )
        # These lines are important when choice of label data
        # results in multiple protobuf nodes being the same in the
        # loaded DAG.
        this_node = node_to_node_d.get(this_node, this_node)
        node_to_node_d[this_node] = this_node

        # Update node_id dictionary
        node_id_to_node[node_id] = this_node

        # Add child edges of this_node
        for child in children:
            this_node.add_edge(child, weight=1, prob=1, prob_norm=False)

    # Last node in postorder should be UA node
    assert this_node.is_ua_node()
    dag = HistoryDag(this_node)
    return pbdag.return_type.from_history_dag(dag)


def load_MAD_protobuf_file(filename, **kwargs):
    """Load a mutation annotated DAG protobuf file and return a
    CGHistoryDag."""
    with open(filename, "rb") as fh:
        pb_data = dpb.data()
        pb_data.ParseFromString(fh.read())
    return load_MAD_protobuf(pb_data, **kwargs)



from frozendict import frozendict

from historydag import HistoryDag
import historydag.parsimony_utils as parsimony_utils
from historydag.utils import UALabel


class SequenceHistoryDag(HistoryDag):
    """A HistoryDag subclass with node labels containing full, unambiguous
    nucleotide sequences.

    The constructor for this class requires that each node label contain a 'sequence'
    field, which is expected to hold an unambiguous nucleotide sequence.

    A HistoryDag containing 'compact_genome' node label fields which contain
    :class:`compact_genome.CompactGenome` objects may be automatically converted to
    this subclass by calling the class method :meth:`SequenceHistoryDag.from_dag`, providing the
    HistoryDag object to be converted.
    """

    _required_label_fields = {
        "sequence": [
            (("compact_genome",), lambda node: node.label.compact_genome.to_sequence())
        ]
    }

    _default_args = frozendict(parsimony_utils.hamming_distance_countfuncs) | {
        "start_func": (lambda n: 0),
        "edge_func": lambda l1, l2: (
            0
            if isinstance(l1, UALabel)
            else parsimony_utils.default_nt_transitions.weighted_hamming_distance(
                l1.sequence, l2.sequence
            )
        ),
        "expand_func": parsimony_utils.default_nt_transitions.ambiguity_map.get_sequence_resolution_func(
            "sequence"
        ),
        "optimal_func": min,
    }

    def hamming_parsimony_count(self):
        """Count the hamming parsimony scores of all trees in the history DAG.

        Returns a Counter with integer keys.
        """
        return self.weight_count(**parsimony_utils.hamming_distance_countfuncs)

    def summary(self):
        HistoryDag.summary(self)
        min_pars, max_pars = self.weight_range_annotate(
            **parsimony_utils.hamming_distance_countfuncs
        )
        print(f"Parsimony score range {min_pars} to {max_pars}")


class AmbiguousLeafSequenceHistoryDag(SequenceHistoryDag):
    """A HistoryDag subclass with node labels containing full nucleotide
    sequences.

    The constructor for this class requires that each node label contain a 'sequence'
    field, which is expected to hold an unambiguous nucleotide sequence if the node
    is internal. The sequence may be ambiguous for leaf nodes.

    A HistoryDag containing 'compact_genome' node label fields which contain
    :class:`compact_genome.CompactGenome` objects may be automatically converted to
    this subclass by calling the class method :meth:`SequenceHistoryDag.from_dag`, providing the
    HistoryDag object to be converted.
    """

    _default_args = frozendict(
        parsimony_utils.leaf_ambiguous_hamming_distance_countfuncs
    ) | {
        "start_func": (lambda n: 0),
        "edge_func": lambda l1, l2: (
            0
            if isinstance(l1, UALabel)
            else parsimony_utils.default_nt_transitions.weighted_hamming_distance(
                l1.sequence, l2.sequence
            )
        ),
        "expand_func": parsimony_utils.default_nt_transitions.ambiguity_map.get_sequence_resolution_func(
            "sequence"
        ),
        "optimal_func": min,
    }

    def hamming_parsimony_count(self):
        """Count the hamming parsimony scores of all trees in the history DAG.

        Returns a Counter with integer keys.
        """
        return self.weight_count(
            **parsimony_utils.leaf_ambiguous_hamming_distance_countfuncs
        )

    def summary(self):
        HistoryDag.summary(self)
        min_pars, max_pars = self.weight_range_annotate(
            **parsimony_utils.leaf_ambiguous_hamming_distance_countfuncs
        )
        print(f"Parsimony score range {min_pars} to {max_pars}")
